# -*- coding: utf-8 -*-
"""
This file contains the Qudi mapper module.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['Converter', 'Mapper']

from PySide2.QtCore import QCoreApplication
from PySide2.QtCore import QThread
from PySide2.QtCore import QTimer
from PySide2.QtWidgets import QAbstractButton
from PySide2.QtWidgets import QAbstractSlider
from PySide2.QtWidgets import QComboBox
from PySide2.QtWidgets import QDoubleSpinBox
from PySide2.QtWidgets import QLineEdit
from PySide2.QtWidgets import QPlainTextEdit
from PySide2.QtWidgets import QSpinBox

import functools

SUBMIT_POLICY_AUTO = 0
"""automatically submit changes"""
SUBMIT_POLICY_MANUAL = 1
"""wait with submitting changes until submit() is called"""


class Converter:
    """
    Class for converting data between display and storage (i.e. widget and
    model).
    """
    def widget_to_model(self, data):
        """
        Converts data from the format given by the widget to the model data format.

        Parameters
        ----------
        data : object
            Data to be converted.

        Returns
        -------
        object
            Converted data.

        """
        return data

    def model_to_widget(self, data):
        """
        Converts data from the model format to the widget data format.

        Parameters
        ----------
        data : object
            Data to be converted from the model format to the widget format.

        Returns
        -------
        object
            Converted data in the widget format.

        """
        return data


class Mapper:
    """
    The Mapper connects a Qt widget for displaying and editing certain data
    types with a model property or setter and getter functions. The model can
    be e.g. a logic or a hardware module.

    Usage Example:
    ==============

    We assume to have a logic module which is connected to our GUI via a
    connector and we can access it by the `logic_module` variable. We
    further assume that this logic module has a string property called
    `some_value` and a signal `some_value_changed` which is emitted when the
    property is changed programmatically.
    In the GUI module we have defined a QLineEdit, e.g. by
    ```
    lineedit = QLineEdit()
    ```
    In the on_activate method of the GUI module, we define the following
    mapping between the line edit and the logic property:
    ```
    def on_activate(self):
        self.mapper = Mapper()
        self.mapper.add_mapping(self.lineedit, self.logic_module,
                'some_value', 'some_value_changed')
    ```
    Now, if the user changes the string in the lineedit, the property of the
    logic module is changed. If the logic module's property is changed
    programmatically, the change is automatically displayed in the GUI.

    If the GUI module is deactivated we should delete all mappings:
    ```
    def on_deactivate(self):
        self.mapper.clear_mapping()
    ```
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._submit_policy = SUBMIT_POLICY_AUTO
        self._mappings = {}

    def _get_property_from_widget(self, widget):
        """
        Returns the property name we determined from the widget's type.
        """
        if isinstance(widget, QAbstractButton):
            return 'checked'
        elif isinstance(widget, QComboBox):
            return 'currentIndex'
        elif isinstance(widget, QLineEdit):
            return 'text'
        elif (isinstance(widget, (QSpinBox,
                                  QDoubleSpinBox,
                                  QAbstractSlider))):
            return 'value'
        elif isinstance(widget, QPlainTextEdit):
            return 'plainText'
        else:
            raise TypeError(f'Property of widget {repr(widget)} could not be determined.')

    def add_mapping(self,
                    widget,
                    model,
                    model_getter,
                    model_property_notifier=None,
                    model_setter=None,
                    widget_property_name='',
                    widget_property_notifier=None,
                    converter=None):
        """
        Adds a mapping.

        Parameters
        ----------
        widget : QtWidget
            A widget displaying some data. You want to map this widget to model data.
        model : object
            Instance of a class holding model data (e.g. a logic or hardware module).
        model_getter : property/callable 
            Either a property holding the data to be displayed in widget or a getter method to retrieve data from the
            model was changed.
        model_property_notifier : Signal
            A signal that is fired when the data was changed. If None then data changes are not monitored and the
            widget is not updated. Default is None.
        model_setter : callable
            A setter method which is called to set data to the model. If model_getter is a property the setter can be
            determined from this property and model_setter is ignored if it is None. If it is not None always this
            callable is used. Default is None.
        widget_property_name : str
            The name of the pyqtProperty of the widget used to map the data. If it is an empty string the relevant
            property is guessed from the widget's type. Default is ''.
        widget_property_notifier : Signal
            Notifier signal which is fired by the widget when the data changed. If None, this is determined directly
            from the property. Example usage: QLineEdit().editingFinished. Default is None.
        converter : Converter
            Converter instance for converting data between widget display and model. Default is None.
        """
        # guess widget property if not specified
        if widget_property_name == '':
            widget_property_name = self._get_property_from_widget(widget)

        # define key of mapping
        key = (widget, widget_property_name)

        # check if already exists
        if key in self._mappings:
            raise RuntimeError(
                f'Property {widget_property_name} of widget {repr(widget)} already mapped.'
            )

        # check if widget property is available
        index = widget.metaObject().indexOfProperty(widget_property_name)
        if index == -1:
            raise RuntimeError(
                f'Property "{widget_property_name}" of widget "{widget.__class__.__name__}" not '
                f'available.'
            )

        meta_property = widget.metaObject().property(index)

        # widget property notifier
        if widget_property_notifier is None:
            # check that widget property as a notify signal
            if not meta_property.hasNotifySignal():
                raise RuntimeError(
                    f'Property "{widget_property_name}" of widget "{widget.__class__.__name__}" '
                    f'has no notify signal.'
                )

            widget_property_notifier = getattr(
                widget,
                meta_property.notifySignal().name().data().decode('utf8'))

        # check that widget property is readable
        if not meta_property.isReadable():
            raise RuntimeError(
                f'Property "{widget_property_name}" of widget "{widget.__class__.__name__}" is not '
                f'readable.'
            )
        widget_property_getter = meta_property.read
        # check that widget property is writable if requested
        if not meta_property.isWritable():
            raise RuntimeError(
                f'Property "{widget_property_name}" of widget "{widget.__class__.__name__}" is not '
                f'writable.'
            )
        widget_property_setter = meta_property.write

        if isinstance(model_getter, str):
            # check if it is a property
            attr = getattr(model.__class__, model_getter, None)
            if attr is None:
                raise AttributeError(f'Model has no attribute "{model_getter}"')
            if isinstance(attr, property):
                # retrieve getter from property
                model_property_name = model_getter
                model_getter = functools.partial(attr.fget, model)
                # if no setter was specified, get it from the property
                if model_setter is None:
                    model_setter = functools.partial(attr.fset, model)
                    if model_getter is None:
                        raise AttributeError(
                            f'Attribute "{model_property_name}" of model is readonly.'
                        )
            else:
                # getter is not a property. Check if it is a callable.
                model_getter_name = model_getter
                model_getter = getattr(model, model_getter)
                if not callable(model_getter):
                    raise AttributeError(
                        f'Attribute "{model_getter_name}" of model is not callable.'
                    )
        if isinstance(model_setter, str):
            model_setter_name = model_setter
            model_setter = getattr(model, model_setter)
            if not callable(model_setter):
                raise AttributeError(f'Attribute "{model_setter_name}" of model is not callable')
        if isinstance(model_property_notifier, str):
            model_property_notifier = getattr(model, model_property_notifier)

        # connect to widget property notifier
        widget_property_notifier_slot = functools.partial(
            self._on_widget_property_notification, key)
        widget_property_notifier.connect(widget_property_notifier_slot)

        # if model_notify_signal was specified, connect to it
        model_property_notifier_slot = None
        if model_property_notifier is not None:
            model_property_notifier_slot = functools.partial(
                self._on_model_notification, key)
            model_property_notifier.connect(model_property_notifier_slot)
        # save mapping
        self._mappings[key] = {
            'widget_property_name': widget_property_name,
            'widget_property_getter': widget_property_getter,
            'widget_property_setter': widget_property_setter,
            'widget_property_notifier': widget_property_notifier,
            'widget_property_notifier_slot': widget_property_notifier_slot,
            'widget_property_notifications_disabled': False,
            'model': model,
            'model_property_setter': model_setter,
            'model_property_getter': model_getter,
            'model_property_notifier': model_property_notifier,
            'model_property_notifier_slot': model_property_notifier_slot,
            'model_property_notifications_disabled': False,
            'converter': converter}

    def _on_widget_property_notification(self, key, *args):
        """
        Event handler for widget property change notification. Used with
        functools.partial to get the widget as first parameter.

        Parameters
        ----------
        key : (QtWidget, str)
            The key consisting of widget and property name, the notification signal was emitted from.
        args*: list
            List of event parameters.
        """
        widget, widget_property_name = key
        if self._mappings[key]['widget_property_notifications_disabled']:
            return
        if self._submit_policy == SUBMIT_POLICY_AUTO:
            self._mappings[key][
                'model_property_notifications_disabled'] = True
            try:
                # get value
                value = self._mappings[key]['widget_property_getter'](
                    widget)
                # convert it if requested
                if self._mappings[key]['converter'] is not None:
                    value = self._mappings[key][
                        'converter'].widget_to_model(value)
                # set it to model
                self._mappings[key]['model_property_setter'](value)
            finally:
                self._mappings[key][
                    'model_property_notifications_disabled'] = False
        else:
            pass

    def _on_model_notification(self, key, *args):
        """
        Event handler for model data change notification. Used with
        functools.partial to get the widget as first parameter.

        Parameters
        ----------
        key : (QtWidget, str)
            The key consisting of widget and property name the notification signal was emitted from.
        args* : list
            List of event parameters
        """
        widget, widget_property_name = key
        mapping = self._mappings[key]
        # get value from model
        value = self._mappings[key]['model_property_getter']()

        # are updates disabled?
        if self._mappings[key]['model_property_notifications_disabled']:
            # but check if value has changed first
            # get value from widget
            value_widget = self._mappings[key]['widget_property_getter'](
                widget)
            # convert it if requested
            if self._mappings[key]['converter'] is not None:
                value_widget = self._mappings[key][
                    'converter'].widget_to_model(value_widget)
            # accept changes, stop if nothing has changed
            if value == value_widget:
                return

        # convert value if requested
        if self._mappings[key]['converter'] is not None:
            value = self._mappings[key]['converter'].model_to_widget(value)

        # update widget
        self._mappings[key][
            'widget_property_notifications_disabled'] = True
        try:
            self._mappings[key]['widget_property_setter'](widget, value)
        finally:
            self._mappings[key][
                'widget_property_notifications_disabled'] = False

    def clear_mapping(self):
        """
        Clears all mappings.
        """
        # convert iterator to list because the _mappings dictionary will
        # change its size during iteration
        for key in list(self._mappings.keys()):
            self.remove_mapping(key)

    def remove_mapping(self, widget, widget_property_name=''):
        """
        Removes the mapping which maps the QtWidget widget to some model data.

        Parameters
        ----------
        widget : QtWidget/(QtWidget, str)
            Widget the mapping is attached to or a tuple containing the widget and the widget's property name.
        widget_property_name : str
            Name of the property of the widget we are dealing with. If '' it will be determined from the widget.
            Default is ''.
        """
        if isinstance(widget, tuple):
            widget, widget_property_name = widget
        # guess widget property if not specified
        if widget_property_name == '':
            widget_property_name = self._get_property_from_widget(widget)
        # define key
        key = (widget, widget_property_name)
        # check that key has a mapping
        if not key in self._mappings:
            raise RuntimeError(f'Widget "{repr(widget)}" is not mapped.')
        # disconnect signals
        self._mappings[key]['widget_property_notifier'].disconnect(
            self._mappings[key]['widget_property_notifier_slot'])
        if self._mappings[key]['model_property_notifier'] is not None:
            self._mappings[key]['model_property_notifier'].disconnect(
                self._mappings[key]['model_property_notifier_slot'])
        # remove from dictionary
        del self._mappings[key]

    @property
    def submit_policy(self):
        """
        Returns the submit policy.
        """
        return self._submit_policy

    @submit_policy.setter
    def submit_policy(self, policy):
        """
        Sets submit policy.

        Submit policy can either be SUBMIT_POLICY_AUTO or
        SUBMIT_POLICY_MANUAL. If the submit policy is auto then changes in
        the widgets are automatically submitted to the model. If manual
        call submit() to submit it.

        Parameters
        ----------
        policy : enum 
            Submit policy.
        """
        if policy not in [SUBMIT_POLICY_AUTO, SUBMIT_POLICY_MANUAL]:
            raise ValueError(f'Unknown submit policy "{policy}"')
        self._submit_policy = policy

    def submit(self):
        """
        Submits the current values stored in the widgets to the models.
        """
        # make sure it is called from main thread
        if (not QThread.currentThread() == QCoreApplication.instance(
        ).thread()):
            QTimer.singleShot(0, self.submit)
            return

        submit_policy = self._submit_policy
        self.submit_policy = SUBMIT_POLICY_AUTO
        try:
            for key in self._mappings:
                self._on_widget_property_notification(key)
        finally:
            self.submit_policy = submit_policy

    def revert(self):
        """
        Takes the data stored in the models and displays them in the widgets.
        """
        # make sure it is called from main thread
        if (not QThread.currentThread() == QCoreApplication.instance(
        ).thread()):
            QTimer.singleShot(0, self.revert)
            return

        for key in self._mappings:
            self._on_model_notification(key)



# -*- coding: utf-8 -*-
"""
This file contains the Qudi Manager class.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

import os
import importlib
import copy
import weakref
import fysom

from typing import FrozenSet, Iterable
from functools import partial
from PySide2 import QtCore

from qudi.util.mutex import RecursiveMutex   # provides access serialization between threads
from qudi.core.logger import get_logger
from qudi.core.servers import get_remote_module_instance
from qudi.core.module import Base, get_module_app_data_path

logger = get_logger(__name__)


class ModuleManager(QtCore.QObject):
    """
    """
    _instance = None  # Only class instance created will be stored here as weakref
    _lock = RecursiveMutex()

    sigModuleStateChanged = QtCore.Signal(str, str, str)
    sigModuleAppDataChanged = QtCore.Signal(str, str, bool)
    sigManagedModulesChanged = QtCore.Signal(dict)

    def __new__(cls, *args, **kwargs):
        with cls._lock:
            if cls._instance is None or cls._instance() is None:
                obj = super().__new__(cls, *args, **kwargs)
                cls._instance = weakref.ref(obj)
                return obj
            raise RuntimeError(
                'ModuleManager is a singleton. An instance has already been created in this '
                'process. Please use ModuleManager.instance() instead.'
            )

    def __init__(self, *args, qudi_main, **kwargs):
        super().__init__(*args, **kwargs)
        self._qudi_main_ref = weakref.ref(qudi_main, self._qudi_main_ref_dead_callback)
        self._modules = dict()

    @classmethod
    def instance(cls):
        with cls._lock:
            if cls._instance is None:
                return None
            return cls._instance()

    def __len__(self):
        with self._lock:
            return len(self._modules)

    def __getitem__(self, key):
        with self._lock:
            return self._modules.__getitem__(key)

    def __setitem__(self, key, value):
        with self._lock:
            if value.name != key:
                raise NameError('ManagedModule.name attribute does not match key')
            self.add_module(value, allow_overwrite=True)

    def __delitem__(self, key):
        self.remove_module(key)

    def __contains__(self, item):
        with self._lock:
            return self._modules.__contains__(item)

    def clear(self):
        with self._lock:
            for module_name in tuple(self._modules):
                self.remove_module(module_name, ignore_missing=True, emit_change=False)
            self.sigManagedModulesChanged.emit(self.modules)

    def get(self, *args):
        with self._lock:
            return self._modules.get(*args)

    def items(self):
        return self._modules.copy().items()

    def values(self):
        return self._modules.copy().values()

    def keys(self):
        return self._modules.copy().keys()

    @property
    def module_names(self):
        with self._lock:
            return tuple(self._modules)

    @property
    def module_states(self):
        with self._lock:
            return {name: mod.state for name, mod in self._modules.items()}

    @property
    def module_instances(self):
        with self._lock:
            return {name: mod.instance for name, mod in self._modules.items() if
                    mod.instance is not None}

    @property
    def modules(self):
        return self._modules.copy()

    def remove_module(self, module_name, ignore_missing=False, emit_change=True):
        with self._lock:
            module = self._modules.pop(module_name, None)
            if module is None and not ignore_missing:
                raise KeyError(f'No module with name "{module_name}" registered.')
            module.deactivate()
            module.sigStateChanged.disconnect(self.sigModuleStateChanged)
            module.sigAppDataChanged.disconnect(self.sigModuleAppDataChanged)
            if module.allow_remote_access:
                remote_modules_server = self._qudi_main_ref().remote_modules_server
                if remote_modules_server is not None:
                    remote_modules_server.remove_shared_module(module_name)
            self.refresh_module_links()
            if emit_change:
                self.sigManagedModulesChanged.emit(self.modules)

    def add_module(self, name, base, configuration, allow_overwrite=False, emit_change=True):
        with self._lock:
            if not isinstance(name, str) or not name:
                raise TypeError('module name must be non-empty str type')
            if base not in ('gui', 'logic', 'hardware'):
                raise ValueError(f'No valid module base "{base}". '
                                 f'Unable to create qudi module "{name}".')
            if allow_overwrite:
                self.remove_module(name, ignore_missing=True)
            elif name in self._modules:
                raise ValueError(f'Module with name "{name}" already registered.')
            module = ManagedModule(self._qudi_main_ref, name, base, configuration)
            module.sigStateChanged.connect(self.sigModuleStateChanged)
            module.sigAppDataChanged.connect(self.sigModuleAppDataChanged)
            self._modules[name] = module
            self.refresh_module_links()
            # Register module in remote module service if module should be shared
            if module.allow_remote_access:
                remote_modules_server = self._qudi_main_ref().remote_modules_server
                if remote_modules_server is None:
                    raise RuntimeError(
                        f'Unable to share qudi module "{module.name}" as remote module. No remote '
                        f'module server running in this qudi process.'
                    )
                else:
                    logger.info(
                        f'Start sharing qudi module "{module.name}" via remote module server.'
                    )
                    remote_modules_server.share_module(module)
            if emit_change:
                self.sigManagedModulesChanged.emit(self.modules)

    def refresh_module_links(self):
        with self._lock:
            weak_refs = {
                name: weakref.ref(mod, partial(self._module_ref_dead_callback, module_name=name))
                for name, mod in self._modules.items()
            }
            for module_name, module in self._modules.items():
                # Add required module references
                required = set(module.connection_cfg.values())
                module.required_modules = set(
                    mod_ref for name, mod_ref in weak_refs.items() if name in required)
                # Add dependent module references
                module.dependent_modules = set(mod_ref for mod_ref in weak_refs.values() if
                                               module_name in mod_ref().connection_cfg.values())

    def activate_module(self, module_name):
        if QtCore.QThread.currentThread() is not self.thread():
            self.current_module_name = module_name
            QtCore.QMetaObject.invokeMethod(self,
            "_activate_module_slot",
            QtCore.Qt.BlockingQueuedConnection)
            return

        with self._lock:
            if module_name not in self._modules:
                raise KeyError(f'No module named "{module_name}" found in managed qudi modules. '
                               f'Module activation aborted.')
            self._modules[module_name].activate()

    def deactivate_module(self, module_name):
        if QtCore.QThread.currentThread() is not self.thread():
            self.current_module_name = module_name
            QtCore.QMetaObject.invokeMethod(self,
            "_deactivate_module_slot",
            QtCore.Qt.BlockingQueuedConnection)
            return

        with self._lock:
            if module_name not in self._modules:
                raise KeyError(f'No module named "{module_name}" found in managed qudi modules. '
                               f'Module deactivation aborted.')
            self._modules[module_name].deactivate()

    def reload_module(self, module_name):
        with self._lock:
            if module_name not in self._modules:
                raise KeyError(f'No module named "{module_name}" found in managed qudi modules. '
                               f'Module reload aborted.')
            return self._modules[module_name].reload()

    def clear_module_app_data(self, module_name):
        with self._lock:
            if module_name not in self._modules:
                raise KeyError(f'No module named "{module_name}" found in managed qudi modules. '
                               f'Can not clear module app status.')
            return self._modules[module_name].clear_module_app_data()

    def has_app_data(self, module_name):
        with self._lock:
            if module_name not in self._modules:
                raise KeyError(f'No module named "{module_name}" found in managed qudi modules. '
                               f'Can not check for app status file.')
            return self._modules[module_name].has_app_data()

    def start_all_modules(self):
        with self._lock:
            for module in self._modules.values():
                module.activate()

    def stop_all_modules(self):
        with self._lock:
            for module in self._modules.values():
                module.deactivate()

    def _module_ref_dead_callback(self, dead_ref, module_name):
        self.remove_module(module_name, ignore_missing=True)

    def _qudi_main_ref_dead_callback(self):
        logger.error('Qudi main reference no longer valid. This should never happen. Tearing down '
                     'ModuleManager.')
        self.clear()

    @QtCore.Slot()
    def _activate_module_slot(self):
        """
        Helper slot that should only be called by activate_module when this method is switching to the main thread.
        """
        self.activate_module(self.current_module_name)

    @QtCore.Slot()
    def _deactivate_module_slot(self):
        """
        Helper slot that should only be called by deactivate_module when this method is switching to the main thread.
        """
        self.deactivate_module(self.current_module_name)


class ManagedModule(QtCore.QObject):
    """Object representing a qudi module (gui, logic or hardware) to be managed by the qudi Manager
     object. Contains status properties and handles initialization, state transitions and
     connection of the module.
    """
    sigStateChanged = QtCore.Signal(str, str, str)
    sigAppDataChanged = QtCore.Signal(str, str, bool)

    _lock = RecursiveMutex()  # Single mutex shared across all ManagedModule instances

    __state_poll_interval = 1  # Max interval in seconds to poll module_state of remote modules

    def __init__(self, qudi_main_ref, name, base, configuration):
        if not isinstance(qudi_main_ref, weakref.ref):
            raise TypeError('qudi_main_ref must be weakref to qudi main instance.')
        if not name or not isinstance(name, str):
            raise ValueError('Module name must be a non-empty string.')
        if base not in ('gui', 'logic', 'hardware'):
            raise ValueError('Module base must be one of ("gui", "logic", "hardware").')

        super().__init__()
        if self.thread() is not QtCore.QCoreApplication.instance().thread():
            raise RuntimeError('ManagedModules can only be owned by the application main thread.')

        self._qudi_main_ref = qudi_main_ref  # Weak reference to qudi main instance
        self._name = name  # Each qudi module needs a unique string identifier
        self._base = base  # Remember qudi module base
        self._instance = None  # Store the module instance later on

        cfg = copy.deepcopy(configuration)

        # Extract module and class name
        self._module, self._class = cfg.get(
            'module.Class',
            'REMOTE.REMOTE'
        ).rsplit('.', 1)
        # Remember connections by name
        self._connect_cfg = cfg.get('connect', dict())
        # See if remotemodules access to this module is allowed
        self._allow_remote_access = cfg.get('allow_remote', False)
        # Extract remote modules URL and certificate if this module is run on a remote machine
        self._remote_module_name = cfg.get('native_module_name', None)
        self._remote_address = cfg.get('address', None)
        self._remote_port = cfg.get('port', None)
        self._remote_certfile = cfg.get('certfile', None)
        self._remote_keyfile = cfg.get('keyfile', None)
        if any(attr is None for attr in [self._remote_module_name, self._remote_address, self._remote_port]):
            self._remote_url = None
        else:
            self._remote_url = f'rpyc://{self._remote_address}:{self._remote_port:d}/{self._remote_module_name}/'
            # Do not propagate remotemodules access
            self._allow_remote_access = False

        # The rest are config options
        self._options = cfg.get('options', dict())

        self._required_modules = frozenset()
        self._dependent_modules = frozenset()

        self.__poll_timer = None
        self.__last_state = None

    def __call__(self):
        return self.instance

    @property
    def name(self):
        return self._name

    @property
    def module_base(self):
        return self._base

    @property
    def class_name(self):
        return self._class

    @property
    def module_name(self):
        return self._module

    @property
    def options(self):
        return copy.deepcopy(self._options)

    @property
    def instance(self):
        with self._lock:
            return self._instance

    @property
    def status_file_path(self):
        return get_module_app_data_path(self.class_name, self.module_base, self.name)

    @property
    def is_loaded(self):
        with self._lock:
            return self._instance is not None

    @property
    def is_active(self):
        with self._lock:
            return self._instance is not None and self._instance.module_state() != 'deactivated'

    @property
    def is_busy(self):
        with self._lock:
            return self.is_active and self._instance.module_state() != 'idle'

    @property
    def is_remote(self):
        return bool(self._remote_url)

    @property
    def allow_remote_access(self):
        return self._allow_remote_access

    @property
    def remote_url(self):
        return self._remote_url

    @property
    def remote_key_path(self):
        return self._remote_keyfile

    @property
    def remote_cert_path(self):
        return self._remote_certfile

    @property
    def state(self):
        try:
            return self._instance.module_state()
        except AttributeError:
            return 'not loaded'
        except:
            return 'BROKEN'

    @property
    def connection_cfg(self):
        return self._connect_cfg.copy()

    @property
    def required_modules(self) -> FrozenSet[weakref.ref]:
        return self._required_modules

    @required_modules.setter
    def required_modules(self, module_iter):
        for module in module_iter:
            if not isinstance(module, weakref.ref):
                raise TypeError('items in required_modules must be weakref.ref instances.')
            if not isinstance(module(), ManagedModule):
                if module() is None:
                    logger.error(
                        f'Dead weakref passed as required module to ManagedModule "{self._name}"'
                    )
                    return
                raise TypeError('required_modules must be iterable of ManagedModule instances '
                                '(or weakref to same instances)')
        self._required_modules = frozenset(module_iter)

    @property
    def dependent_modules(self):
        return self._dependent_modules

    @dependent_modules.setter
    def dependent_modules(self, module_iter):
        dep_modules = set()
        for module in module_iter:
            if not isinstance(module, weakref.ref):
                raise TypeError('items in dependent_modules must be weakref.ref instances.')
            if not isinstance(module(), ManagedModule):
                if module() is None:
                    logger.error(
                        f'Dead weakref passed as dependent module to ManagedModule "{self._name}"'
                    )
                    return
                raise TypeError('dependent_modules must be iterable of ManagedModule instances '
                                '(or weakref to same instances)')
            dep_modules.add(module)
        self._dependent_modules = frozenset(dep_modules)

    @property
    def ranking_active_dependent_modules(self):
        with self._lock:
            active_dependent_modules = set()
            for module_ref in self.dependent_modules:
                module = module_ref()
                if module is None:
                    logger.warning(f'Dead dependent module weakref encountered in ManagedModule '
                                   f'"{self._name}".')
                    continue
                if module.is_active:
                    active_modules = module.ranking_active_dependent_modules
                    if active_modules:
                        active_dependent_modules.update(active_modules)
                    else:
                        active_dependent_modules.add(module_ref)
            return active_dependent_modules

    @property
    def module_thread_name(self):
        return f'mod-{self._base}-{self._name}'

    @property
    def has_app_data(self):
        with self._lock:
            return os.path.exists(self.status_file_path)

    @QtCore.Slot()
    def clear_module_app_data(self):
        with self._lock:
            try:
                os.remove(self.status_file_path)
            except OSError:
                pass
            finally:
                self.sigAppDataChanged.emit(self._base, self._name, self.has_app_data)

    @QtCore.Slot()
    def activate(self) -> None:
        # Switch to the main thread if this method was called from another thread
        if QtCore.QThread.currentThread() is not self.thread():
            QtCore.QMetaObject.invokeMethod(self, 'activate', QtCore.Qt.BlockingQueuedConnection)
            if not self.is_active:
                raise RuntimeError(f'Failed to activate {self.module_base} module "{self.name}"!')
            return

        with self._lock:
            if not self.is_loaded:
                self._load()

            if self.is_remote and self.__poll_timer is None:
                self.__poll_timer = QtCore.QTimer(self)
                logger.debug(f"creating new timer {self.__poll_timer}")
                self.__poll_timer.setInterval(int(round(self.__state_poll_interval * 1000)))
                self.__poll_timer.setSingleShot(True)
                self.__poll_timer.timeout.connect(self._poll_module_state)
                self.__poll_timer.start()

            # Return early if already active
            if self.is_active:
                # If it is a GUI module, show it again.
                if self.module_base == 'gui':
                    self._instance.show()
                return

            if self.is_remote:
                logger.info(f'Activating remote {self.module_base} module "{self.remote_url}"')
            else:
                # this is a local module and not already active
                logger.info(
                    f'Activating {self.module_base} module "{self.module_name}.{self.class_name}"'
                )
                self._instance.module_state.sigStateChanged.connect(self._state_change_callback)

            # Recursive activation of required modules
            for module_ref in self.required_modules:
                module = module_ref()
                if module is None:
                    raise ReferenceError(f'Dead required module weakref encountered in '
                                         f'ManagedModule "{self._name}".')
                module.activate()

            # Establish module interconnections via Connector meta object in qudi module instance
            self._connect()

            # Activate this module
            if self._instance.is_module_threaded:
                thread_name = self.module_thread_name
                thread_manager = self._qudi_main_ref().thread_manager
                thread = thread_manager.get_new_thread(thread_name)
                self._instance.moveToThread(thread)
                thread.start()
                try:
                    QtCore.QMetaObject.invokeMethod(self._instance.module_state,
                                                    'activate',
                                                    QtCore.Qt.BlockingQueuedConnection)
                finally:
                    # Cleanup if activation was not successful
                    if not self.is_active:
                        QtCore.QMetaObject.invokeMethod(self._instance,
                                                        'move_to_main_thread',
                                                        QtCore.Qt.BlockingQueuedConnection)
                        thread_manager.quit_thread(thread_name)
                        thread_manager.join_thread(thread_name)
                        self._disconnect()
                        self._disable_state_updated()

            else:
                try:
                    self._instance.module_state.activate()
                except Exception as e:
                    raise RuntimeError(f'Failed to activate {self.module_base} module "{self.name}"!') from e
                finally:  # cleanup if activation was not successful
                    if not self.is_active:
                        self._disconnect()
                        self._disable_state_updated()

            self.__last_state = self.state

            self.sigAppDataChanged.emit(self._base, self._name, self.has_app_data)

            # Raise exception if by some reason no exception propagated to here and the activation
            # is still unsuccessful.
            if not self.is_active:
                try:
                    self._disconnect()
                    self._disable_state_updated()
                except:
                    pass
                raise RuntimeError(f'Failed to activate {self.module_base} module "{self.name}"!')

    @QtCore.Slot()
    def _poll_module_state(self):
        with self._lock:
            state = self.state
            if state != self.__last_state:
                self.__last_state = state
                self.sigStateChanged.emit(self._base, self._name, state)
            try:
                self.__poll_timer.start()
            except AttributeError:
                pass

    @QtCore.Slot(object)
    def _state_change_callback(self, event=None):
        self.sigStateChanged.emit(self._base, self._name, self.state)

    @QtCore.Slot()
    def deactivate(self):
        # Switch to the main thread if this method was called from another thread
        if QtCore.QThread.currentThread() is not self.thread():
            QtCore.QMetaObject.invokeMethod(self, 'deactivate', QtCore.Qt.BlockingQueuedConnection)
            if self.is_active:
                raise RuntimeError(f'Failed to deactivate {self.module_base} module "{self.name}"!')
            return

        with self._lock:
            if not self.is_active:
                return

            if self.is_remote:
                logger.info(f'Deactivating remote {self.module_base} module "{self.remote_url}"')
            else:
                logger.info(
                    f'Deactivating {self.module_base} module "{self.module_name}.{self.class_name}"'
                )

            # Recursively deactivate dependent modules
            for module_ref in self.dependent_modules:
                module = module_ref()
                if module is None:
                    raise ReferenceError(
                        f'Dead dependent module weakref encountered in ManagedModule "{self.name}".'
                    )
                module.deactivate()

            self._disable_state_updated()

            # Actual deactivation of this module
            if self._instance.is_module_threaded:
                thread_name = self.module_thread_name
                thread_manager = self._qudi_main_ref().thread_manager
                try:
                    QtCore.QMetaObject.invokeMethod(self._instance.module_state,
                                                    'deactivate',
                                                    QtCore.Qt.BlockingQueuedConnection)
                finally:
                    QtCore.QMetaObject.invokeMethod(self._instance,
                                                    'move_to_main_thread',
                                                    QtCore.Qt.BlockingQueuedConnection)
                    thread_manager.quit_thread(thread_name)
                    thread_manager.join_thread(thread_name)
            else:
                try:
                    self._instance.module_state.deactivate()
                except fysom.Canceled:
                    pass
            QtCore.QCoreApplication.instance().processEvents()  # ToDo: Is this still needed?

            # Disconnect modules from this module
            self._disconnect()

            self.__last_state = self.state
            self.sigStateChanged.emit(self._base, self._name, self.__last_state)
            self.sigAppDataChanged.emit(self._base, self._name, self.has_app_data)

            # Raise exception if by some reason no exception propagated to here and the deactivation
            # is still unsuccessful.
            if self.is_active:
                raise RuntimeError(f'Failed to deactivate {self.module_base} module "{self.name}"!')

    @QtCore.Slot()
    def reload(self):
        # Switch to the main thread if this method was called from another thread
        if QtCore.QThread.currentThread() is not self.thread():
            QtCore.QMetaObject.invokeMethod(self, 'reload', QtCore.Qt.BlockingQueuedConnection)
            return

        with self._lock:
            # Deactivate if active
            was_active = self.is_active
            mod_to_activate = None
            if was_active:
                mod_to_activate = self.ranking_active_dependent_modules
                self.deactivate()

            # reload module
            self._load(reload=True)

            # re-activate all modules that have been active before
            if was_active:
                if mod_to_activate:
                    for module_ref in mod_to_activate:
                        module = module_ref()
                        if module is None:
                            continue
                        module.activate()
                else:
                    self.activate()

    def _load(self, reload=False):
        """
        """
        with self._lock:
            try:
                # Do nothing if already loaded and no reload is requested
                if self.is_loaded and not reload:
                    return

                if self.is_remote:
                    try:
                        self._instance = get_remote_module_instance(self.remote_url,
                                                                    certfile=self._remote_certfile,
                                                                    keyfile=self._remote_keyfile)
                    except BaseException as e:
                        self._instance = None
                        raise RuntimeError(f'Error during initialization of remote '
                                           f'{self.module_base} module {self.remote_url}') from e
                else:
                    # qudi module import and reload
                    mod = importlib.import_module(f'qudi.{self._base}.{self._module}')
                    if reload:
                        importlib.reload(mod)

                    # Try getting qudi module class from imported module
                    mod_class = getattr(mod, self._class, None)
                    if mod_class is None:
                        raise AttributeError(f'No module class "{self._class}" found in module '
                                             f'"qudi.{self._base}.{self._module}"')

                    # Check if imported class is a valid qudi module class
                    if not issubclass(mod_class, Base):
                        raise TypeError(f'Qudi module class "{mod_class}" is no subclass of '
                                        f'"qudi.core.module.Base"')

                    # Try to instantiate the imported qudi module class
                    try:
                        self._instance = mod_class(qudi_main_weakref=self._qudi_main_ref,
                                                   name=self._name,
                                                   config=self._options)
                    except BaseException as e:
                        self._instance = None
                        raise RuntimeError(f'Error during initialization of qudi module '
                                           f'"qudi.{self._base}.{self._module}.{self._class}"')
            finally:
                self.__last_state = self.state
                self.sigStateChanged.emit(self._base, self._name, self.__last_state)

    def _connect(self):
        with self._lock:
            # Check if module has already been loaded/instantiated
            if not self.is_loaded:
                raise RuntimeError(f'Connection failed. No module instance found for module '
                                   f'"{self._base}.{self._name}".')

            # Collect all module instances required by connector config
            module_instances = {
                module_ref().name: module_ref().instance for module_ref in self.required_modules
            }
            module_connections = {conn_name: module_instances[mod_name] for conn_name, mod_name in
                                  self._connect_cfg.items()}

            # Apply module connections
            self._instance.connect_modules(module_connections)

    def _disconnect(self):
        with self._lock:
            self._instance.disconnect_modules()

    def _disable_state_updated(self):
        try:
            self.__poll_timer.stop()
            self.__poll_timer.timeout.disconnect()
        except AttributeError:
            self._instance.module_state.sigStateChanged.disconnect(self._state_change_callback)
        finally:
            self.__poll_timer = None



# -*- coding: utf-8 -*-

"""
This module contains the QTreeWidget object to display qudi configurations.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['ConfigQTreeWidget']

from typing import Any, Mapping
from collections.abc import Sequence as AbstractSequence
from collections.abc import Mapping as AbstractMapping
from PySide2 import QtWidgets


class ConfigQTreeWidget(QtWidgets.QTreeWidget):
    """Specialized QTreeWidget to display qudi configurations.
    """

    def set_config(self, config: Mapping[str, Any]) -> None:
        self.clear()
        self._insert_value(self.invisibleRootItem(), config)

    def _insert_value(self, root: QtWidgets.QTreeWidgetItem, value: Any) -> None:
        """Recursively fill the QTreeWidgeItem.
        """
        # if value is a mapping, open up a new sub-tree and recursively fill it
        if isinstance(value, AbstractMapping):
            if len(value) == 0:
                root.setText(0, f'{root.text(0)} {{}}')
            else:
                for key, val in value.items():
                    child = QtWidgets.QTreeWidgetItem()
                    child.setText(0, f'{key}:')
                    root.addChild(child)
                    self._insert_value(child, val)
                root.setExpanded(True)
        # if value is any string (or similar) type, e.g. str, bytes, bytearray, just add it to root
        elif isinstance(value, (str, bytearray, bytes)):
            try:
                text = value.decode('utf8')
            except (UnicodeDecodeError, AttributeError):
                text = str(value)
            root.setText(0, f'{root.text(0)} "{text}"')
        # If the value is a sequence of values itself, open up a new sub-tree and fill it
        elif isinstance(value, AbstractSequence):
            if len(value) == 0:
                root.setText(0, f'{root.text(0)} []')
            else:
                for val in value:
                    child = QtWidgets.QTreeWidgetItem()
                    child.setText(0, '-')
                    root.addChild(child)
                    self._insert_value(child, val)
                root.setExpanded(True)
        # If the value is anything else, just convert to str
        else:
            root.setText(0, f'{root.text(0)} {str(value)}')



# -*- coding: utf-8 -*-

"""
This file contains unit tests for all qudi fit routines for exponential decay models.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

import unittest
import numpy as np

from qudi.util.fit_models.exp_decay import StretchedExponentialDecay


class TestExpDecayMethods(unittest.TestCase):
    _fit_param_tolerance = 0.05  # 5% tolerance for each fit parameter

    @staticmethod
    def stretched_exp_decay(x, offset, amplitude, decay, stretch):
        return offset + amplitude * np.exp(-(x / decay) ** stretch)

    def setUp(self):
        self.offset = (np.random.rand() - 0.5) * 2e6
        self.amplitude = np.random.rand() * 100
        window = max(1e-9, np.random.rand() * 1e9)
        left = (np.random.rand() - 0.5) * 2e9
        points = np.random.randint(10, 1001)
        self.x_values = np.linspace(left, left + window, points)
        min_decay = 1.5 * (self.x_values[1] - self.x_values[0])
        self.decay = min_decay + np.random.rand() * ((window / 2) - min_decay)
        self.stretch = 0.1 + np.random.rand() * 3.9
        self.noise_amp = max(self.amplitude / 10, np.random.rand() * self.amplitude)
        self.noise = (np.random.rand(points) - 0.5) * self.noise_amp

    def test_exp_decay(self):
        # Test for exponential decay (stretch == 1)
        y_values = self.noise + self.stretched_exp_decay(self.x_values,
                                                         self.offset,
                                                         self.amplitude,
                                                         self.decay,
                                                         1)

        fit_model = StretchedExponentialDecay()
        guess = fit_model.guess(y_values, self.x_values)
        guess['stretch'].set(vary=False, value=1)
        fit_result = fit_model.fit(data=y_values, x=self.x_values, **guess)

        params_ideal = {'offset': self.offset,
                        'amplitude': self.amplitude,
                        'decay': self.decay}
        for name, ideal_val in params_ideal.items():
            diff = abs(fit_result.best_values[name] - ideal_val)
            tolerance = abs(ideal_val * self._fit_param_tolerance)
            msg = 'Exp. decay fit parameter "{0}" not within {1:.2%} tolerance'.format(
                name, self._fit_param_tolerance
            )
            self.assertLessEqual(diff, tolerance, msg)

    def test_stretched_exp_decay(self):
        # Test for stretched exponential decay
        y_values = self.noise + self.stretched_exp_decay(self.x_values,
                                                         self.offset,
                                                         self.amplitude,
                                                         self.decay,
                                                         self.stretch)

        fit_model = StretchedExponentialDecay()
        fit_result = fit_model.fit(data=y_values,
                                   x=self.x_values,
                                   **fit_model.guess(y_values, self.x_values))

        params_ideal = {'offset': self.offset,
                        'amplitude': self.amplitude,
                        'decay': self.decay,
                        'stretch': self.stretch}
        for name, ideal_val in params_ideal.items():
            diff = abs(fit_result.best_values[name] - ideal_val)
            tolerance = abs(ideal_val * self._fit_param_tolerance)
            msg = 'Stretched exp. decay fit parameter "{0}" not within {1:.2%} tolerance'.format(
                name, self._fit_param_tolerance
            )
            self.assertLessEqual(diff, tolerance, msg)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
This file contains the Qudi module base class.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""
import logging
import os
import copy
import uuid
from abc import abstractmethod
from uuid import uuid4
from fysom import Fysom
from PySide2 import QtCore, QtGui, QtWidgets
from typing import Any, Mapping, Optional, Callable, Union, Dict

from qudi.core.configoption import MissingOption
from qudi.core.statusvariable import StatusVar
from qudi.util.paths import get_module_app_data_path, get_daily_directory, get_default_data_dir
from qudi.util.yaml import yaml_load, yaml_dump
from qudi.core.meta import ModuleMeta
from qudi.core.logger import get_logger


class ModuleStateMachine(Fysom, QtCore.QObject):
    """
    FIXME
    """
    # do not copy declaration of trigger(self, event, *args, **kwargs), just apply Slot decorator
    trigger = QtCore.Slot(str, result=bool)(Fysom.trigger)

    # signals
    sigStateChanged = QtCore.Signal(object)  # Fysom event

    def __init__(self, callbacks=None, parent=None, **kwargs):
        if callbacks is None:
            callbacks = dict()

        # State machine definition
        # the abbreviations for the event list are the following:
        #   name:   event name,
        #   src:    source state,
        #   dst:    destination state
        fsm_cfg = {'initial': 'deactivated',
                   'events': [{'name': 'activate', 'src': 'deactivated', 'dst': 'idle'},
                              {'name': 'deactivate', 'src': 'idle', 'dst': 'deactivated'},
                              {'name': 'deactivate', 'src': 'locked', 'dst': 'deactivated'},
                              {'name': 'lock', 'src': 'idle', 'dst': 'locked'},
                              {'name': 'unlock', 'src': 'locked', 'dst': 'idle'}],
                   'callbacks': callbacks}

        # Initialise state machine:
        super().__init__(parent=parent, cfg=fsm_cfg, **kwargs)

    def __call__(self) -> str:
        """
        Returns the current state.
        """
        return self.current

    def on_change_state(self, e: Any) -> None:
        """Fysom callback for all state transitions.

        Parameters
        ----------
        e : object
            Fysom event object passed through all state transition callbacks.
        """
        self.sigStateChanged.emit(e)

    @QtCore.Slot()
    def activate(self) -> None:
        super().activate()

    @QtCore.Slot()
    def deactivate(self) -> None:
        super().deactivate()

    @QtCore.Slot()
    def lock(self) -> None:
        super().lock()

    @QtCore.Slot()
    def unlock(self) -> None:
        super().unlock()


class Base(QtCore.QObject, metaclass=ModuleMeta):
    """Base class for all loadable modules

    * Ensure that the program will not die during the load of modules
    * Initialize modules
    * Provides a self identification of the used module
    * per-module logging facility
    * Provides a self de-initialization of the used module
    * Get your own configuration (for save)
    * Get name of status variables
    * Get status variables
    * Reload module data (from saved variables)
    """
    _threaded = False

    # FIXME: This __new__ implementation has the sole purpose to circumvent a known PySide2(6) bug.
    #  See https://bugreports.qt.io/browse/PYSIDE-1434 for more details.
    def __new__(cls, *args, **kwargs):
        abstract = getattr(cls, '__abstractmethods__', frozenset())
        if abstract:
            raise TypeError(f'Can\'t instantiate abstract class "{cls.__name__}" '
                            f'with abstract methods {set(abstract)}')
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, qudi_main_weakref: Any, name: str,
                 config: Optional[Mapping[str, Any]] = None,
                 callbacks: Optional[Mapping[str, Callable]] = None, **kwargs):
        """Initialise Base instance. Set up its state machine and initialize ConfigOption meta
        attributes from given config.

        Parameters
        ----------
        self : object
            The object being initialized.
        name : str
            Unique name for this module instance.
        configuration : dict
            Parameters from the configuration file.
        callbacks : dict
            Dictionary specifying functions to be run on state machine transitions.
        """
        super().__init__(**kwargs)

        if config is None:
            config = dict()
        if callbacks is None:
            callbacks = dict()

        # Keep weak reference to qudi main instance
        self.__qudi_main_weakref = qudi_main_weakref

        # Create logger instance for module
        self.__logger = get_logger(f'{self.__module__}.{self.__class__.__name__}')

        # Create a copy of the _meta class dict and attach it to the created instance
        self._meta = copy.deepcopy(self._meta)
        # Add additional meta info to _meta dict
        self._meta['name'] = name
        self._meta['uuid'] = uuid4()
        self._meta['configuration'] = copy.deepcopy(config)

        # set instance attributes according to config_option meta objects
        self.__initialize_config_options(config)

        # set instance attributes according to connector meta objects
        self.__initialize_connectors()

        # Initialize module FSM
        default_callbacks = {'on_before_activate'  : self.__activation_callback,
                             'on_before_deactivate': self.__deactivation_callback}
        default_callbacks.update(callbacks)
        self.module_state = ModuleStateMachine(parent=self, callbacks=default_callbacks)
        return

    def __initialize_config_options(self, config: Optional[Mapping[str, Any]]) -> None:
        for attr_name, cfg_opt in self._meta['config_options'].items():
            if cfg_opt.name in config:
                cfg_val = copy.deepcopy(config[cfg_opt.name])
            else:
                if cfg_opt.missing == MissingOption.error:
                    raise ValueError(
                        f'Required ConfigOption "{cfg_opt.name}" not given in configuration.\n'
                        f'Configuration is: {config}'
                    )
                msg = f'No ConfigOption "{cfg_opt.name}" configured, using default value ' \
                      f'"{cfg_opt.default}" instead.'
                cfg_val = copy.deepcopy(cfg_opt.default)
                if cfg_opt.missing == MissingOption.warn:
                    self.log.warning(msg)
                elif cfg_opt.missing == MissingOption.info:
                    self.log.info(msg)
            if cfg_opt.check(cfg_val):
                cfg_val = cfg_opt.convert(cfg_val)
                if cfg_opt.constructor_function is not None:
                    cfg_val = cfg_opt.constructor_function(self, cfg_val)
                setattr(self, attr_name, cfg_val)

    def __initialize_connectors(self) -> None:
        for attr_name, conn in self._meta['connectors'].items():
            setattr(self, attr_name, conn)

    def __eq__(self, other):
        if isinstance(other, Base):
            return self.module_uuid == other.module_uuid
        return super().__eq__(other)

    def __hash__(self):
        return self.module_uuid.int

    @QtCore.Slot()
    def move_to_main_thread(self) -> None:
        """Method that will move this module into the main/manager thread.
        """
        if QtCore.QThread.currentThread() != self.thread():
            QtCore.QMetaObject.invokeMethod(self,
                                            'move_to_main_thread',
                                            QtCore.Qt.BlockingQueuedConnection)
        else:
            self.moveToThread(QtCore.QCoreApplication.instance().thread())

    @property
    def module_thread(self) -> Union[QtCore.QThread, None]:
        """Read-only property returning the current module QThread instance if the module is
        threaded. Returns None otherwise.
        """
        if self._threaded:
            return self.thread()
        return None

    @property
    def module_name(self) -> str:
        """Read-only property returning the module name of this module instance as specified in the
        config.
        """
        return self._meta['name']

    @property
    def module_base(self) -> str:
        """Read-only property returning the module base of this module instance
        ('hardware' 'logic' or 'gui').
        """
        return self._meta['base']

    @property
    def module_uuid(self) -> uuid.UUID:
        """Read-only property returning a unique uuid for this module instance.
        """
        return self._meta['uuid']

    @property
    def module_default_data_dir(self) -> str:
        """Read-only property returning the generic default directory in which to save data.
        Module implementations can overwrite this property with a custom path but should only do so
        with a very good reason.
        """
        config = self._qudi_main.configuration
        data_root = config['default_data_dir']
        if data_root is None:
            data_root = get_default_data_dir()
        if config['daily_data_dirs']:
            data_dir = os.path.join(get_daily_directory(root=data_root), self.module_name)
        else:
            data_dir = os.path.join(data_root, self.module_name)
        return data_dir

    @property
    def module_status_variables(self) -> Dict[str, Any]:
        variables = dict()
        try:
            for attr_name, var in self._meta['status_variables'].items():
                if hasattr(self, attr_name):
                    value = getattr(self, attr_name)
                    if not isinstance(value, StatusVar):
                        if var.representer_function is not None:
                            value = var.representer_function(self, value)
                        variables[var.name] = value
        except:
            self.log.exception('Error while collecting status variables:')
        return variables

    @property
    def _qudi_main(self) -> Any:
        qudi_main = self.__qudi_main_weakref()
        if qudi_main is None:
            raise RuntimeError(
                'Unexpected missing qudi main instance. It has either been deleted or garbage '
                'collected.'
            )
        return qudi_main

    @property
    def log(self) -> logging.Logger:
        """Returns the module logger instance.
        """
        return self.__logger

    @property
    def is_module_threaded(self) -> bool:
        """Returns whether the module shall be started in its own thread.
        """
        return self._threaded

    def __activation_callback(self, event=None) -> bool:
        """Restore status variables before activation and invoke on_activate method.
        """
        try:
            self._load_status_variables()
            self.on_activate()
        except:
            self.log.exception('Exception during activation:')
            return False
        return True

    def __deactivation_callback(self, event=None) -> bool:
        """Invoke on_deactivate method and save status variables afterwards even if deactivation
        fails.
        """
        try:
            self.on_deactivate()
        except:
            self.log.exception('Exception during deactivation:')
        finally:
            # save status variables even if deactivation failed
            self._dump_status_variables()
        return True

    def _load_status_variables(self) -> None:
        """Load status variables from app data directory on disc.
        """
        # Load status variables from app data directory
        file_path = get_module_app_data_path(self.__class__.__name__,
                                             self.module_base,
                                             self.module_name)
        try:
            variables = yaml_load(file_path, ignore_missing=True)
        except:
            variables = dict()
            self.log.exception('Failed to load status variables:')

        # Set instance attributes according to StatusVar meta objects
        try:
            for attr_name, var in self._meta['status_variables'].items():
                value = variables.get(var.name, copy.deepcopy(var.default))
                if var.constructor_function is not None:
                    value = var.constructor_function(self, value)
                setattr(self, attr_name, value)
        except:
            self.log.exception('Error while settings status variables:')

    def _dump_status_variables(self) -> None:
        """Dump status variables to app data directory on disc.

        This method can also be used to manually dump status variables independent of the automatic
        dump during module deactivation.
        """
        file_path = get_module_app_data_path(self.__class__.__name__,
                                             self.module_base,
                                             self.module_name)
        # collect StatusVar values into dictionary
        variables = self.module_status_variables
        # Save to file if any StatusVars have been found
        if variables:
            try:
                yaml_dump(file_path, variables)
            except:
                self.log.exception('Failed to save status variables:')

    def _send_balloon_message(self, title: str, message: str, time: Optional[float] = None,
                              icon: Optional[QtGui.QIcon] = None) -> None:
        qudi_main = self.__qudi_main_weakref()
        if qudi_main is None:
            return
        if qudi_main.gui is None:
            log = get_logger('balloon-message')
            log.warning(f'{title}:\n{message}')
            return
        qudi_main.gui.balloon_message(title, message, time, icon)

    def _send_pop_up_message(self, title: str, message: str):
        qudi_main = self.__qudi_main_weakref()
        if qudi_main is None:
            return
        if qudi_main.gui is None:
            log = get_logger('pop-up-message')
            log.warning(f'{title}:\n{message}')
            return
        qudi_main.gui.pop_up_message(title, message)

    def connect_modules(self, connections: Mapping[str, Any]) -> None:
        """Connects given modules (values) to their respective Connector (keys).

        DO NOT CALL THIS METHOD UNLESS YOU KNOW WHAT YOU ARE DOING!
        """
        # Sanity checks
        conn_names = set(conn.name for conn in self._meta['connectors'].values())
        mandatory_conn = set(
            conn.name for conn in self._meta['connectors'].values() if not conn.optional
        )
        configured_conn = set(connections)
        if not configured_conn.issubset(conn_names):
            raise KeyError(f'Mismatch of connectors in configuration {configured_conn} and module '
                           f'Connector meta objects {conn_names}.')
        if not mandatory_conn.issubset(configured_conn):
            raise ValueError(f'Not all mandatory connectors are specified in config.\n'
                             f'Mandatory connectors are: {mandatory_conn}')

        # Iterate through module connectors and connect them if possible
        for conn in self._meta['connectors'].values():
            target = connections.get(conn.name, None)
            if target is None:
                continue
            if conn.is_connected:
                raise RuntimeError(f'Connector "{conn.name}" already connected.\n'
                                   f'Call "disconnect_modules()" before trying to reconnect.')
            conn.connect(target)

    def disconnect_modules(self) -> None:
        """Disconnects all Connector instances for this module.

        DO NOT CALL THIS METHOD UNLESS YOU KNOW WHAT YOU ARE DOING!
        """
        for conn in self._meta['connectors'].values():
            conn.disconnect()

    @abstractmethod
    def on_activate(self) -> None:
        """Method called when module is activated. Must be implemented by actual qudi module.
        """
        raise NotImplementedError('Please implement and specify the activation method.')

    @abstractmethod
    def on_deactivate(self) -> None:
        """Method called when module is deactivated. Must be implemented by actual qudi module.
        """
        raise NotImplementedError('Please implement and specify the deactivation method.')


class LogicBase(Base):
    """
    """
    _threaded = True


class GuiBase(Base):
    """This is the GUI base class. It provides functions that every GUI module should have.
    """
    _threaded = False
    __window_geometry = StatusVar(name='_GuiBase__window_geometry', default=None)
    __window_state = StatusVar(name='_GuiBase__window_state', default=None)

    @abstractmethod
    def show(self) -> None:
        raise NotImplementedError('Every GUI module needs to implement the show() method!')

    def _save_window_geometry(self, window: QtWidgets.QMainWindow) -> None:
        try:
            self.__window_geometry = window.saveGeometry().toHex().data().decode('utf-8')
        except:
            self.log.exception('Unable to save window geometry:')
            self.__window_geometry = None
        try:
            self.__window_state = window.saveState().toHex().data().decode('utf-8')
        except:
            self.log.exception('Unable to save window geometry:')
            self.__window_state = None

    def _restore_window_geometry(self, window: QtWidgets.QMainWindow) -> bool:
        if isinstance(self.__window_geometry, str):
            try:
                encoded = QtCore.QByteArray(self.__window_geometry.encode('utf-8'))
                window.restoreGeometry(QtCore.QByteArray.fromHex(encoded))
            except:
                self.log.exception('Unable to restore window geometry:')
        if isinstance(self.__window_state, str):
            try:
                encoded = QtCore.QByteArray(self.__window_state.encode('utf-8'))
                return window.restoreState(QtCore.QByteArray.fromHex(encoded))
            except:
                self.log.exception('Unable to restore window state:')
        return False



# -*- coding: utf-8 -*-
"""
This file contains the QMainWindow class for the Qudi main GUI.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

import os
from PySide2 import QtCore, QtGui, QtWidgets
from qudi.core.gui.main_gui.logwidget import LogWidget
from qudi.core.gui.main_gui.remotewidget import RemoteWidget
from qudi.core.gui.main_gui.aboutqudidialog import AboutQudiDialog
from qudi.core.gui.main_gui.settingsdialog import SettingsDialog
from qudi.core.gui.main_gui.modulewidget import ModuleWidget
from qudi.core.gui.main_gui.configwidget import ConfigQTreeWidget
from qudi.util.paths import get_artwork_dir
from qudi.util.widgets.advanced_dockwidget import AdvancedDockWidget
from qtconsole.rich_jupyter_widget import RichJupyterWidget


class QudiMainWindow(QtWidgets.QMainWindow):
    """
    Main Window definition for the manager GUI.
    """
    def __init__(self, parent=None, debug_mode=False, **kwargs):
        super().__init__(parent, **kwargs)
        self.setWindowTitle('qudi: Manager')
        screen_size = QtWidgets.QApplication.instance().primaryScreen().availableSize()
        width = (screen_size.width() * 3) // 4
        height = (screen_size.height() * 3) // 4
        self.resize(width, height)

        self.module_widget = ModuleWidget()
        self.module_widget.setObjectName('moduleTabWidget')
        self.setCentralWidget(self.module_widget)

        # Create actions
        # Toolbar actions
        icon_path = os.path.join(get_artwork_dir(), 'icons')
        self.action_load_configuration = QtWidgets.QAction()
        self.action_load_configuration.setIcon(
            QtGui.QIcon(os.path.join(icon_path, 'document-open')))
        self.action_load_configuration.setText('Load configuration')
        self.action_load_configuration.setToolTip('Load configuration')
        self.action_open_configuration_editor = QtWidgets.QAction()
        self.action_open_configuration_editor.setIcon(
            QtGui.QIcon(os.path.join(icon_path, 'document-new'))
        )
        self.action_open_configuration_editor.setText('New configuration')
        self.action_open_configuration_editor.setToolTip(
            'Open the graphical configuration editor for editing or creating a new qudi '
            'configuration'
        )
        self.action_reload_qudi = QtWidgets.QAction()
        self.action_reload_qudi.setIcon(
            QtGui.QIcon(os.path.join(icon_path, 'view-refresh')))
        self.action_reload_qudi.setText('Reload current configuration')
        self.action_reload_qudi.setToolTip('Reload current configuration')
        self.action_load_all_modules = QtWidgets.QAction()
        self.action_load_all_modules.setIcon(
            QtGui.QIcon(os.path.join(icon_path, 'dialog-warning')))
        self.action_load_all_modules.setText('Load all modules')
        self.action_load_all_modules.setToolTip('Load all available modules found in configuration')
        # quit action
        self.action_quit = QtWidgets.QAction()
        self.action_quit.setIcon(QtGui.QIcon(os.path.join(icon_path, 'application-exit')))
        self.action_quit.setText('Quit qudi')
        self.action_quit.setToolTip('Quit qudi')
        self.action_quit.setShortcut(QtGui.QKeySequence('Ctrl+Q'))
        # view actions
        self.action_view_console = QtWidgets.QAction()
        self.action_view_console.setCheckable(True)
        self.action_view_console.setChecked(True)
        self.action_view_console.setText('Show console')
        self.action_view_console.setToolTip('Show IPython console')
        self.action_view_log = QtWidgets.QAction()
        self.action_view_log.setCheckable(True)
        self.action_view_log.setChecked(True)
        self.action_view_log.setText('Show log')
        self.action_view_log.setToolTip('Show log dockwidget')
        self.action_view_config = QtWidgets.QAction()
        self.action_view_config.setCheckable(True)
        self.action_view_config.setChecked(False)
        self.action_view_config.setText('Show configuration')
        self.action_view_config.setToolTip('Show configuration dockwidget')
        self.action_view_remote = QtWidgets.QAction()
        self.action_view_remote.setCheckable(True)
        self.action_view_remote.setChecked(False)
        self.action_view_remote.setText('Show remote')
        self.action_view_remote.setToolTip('Show remote connections dockwidget')
        self.action_view_threads = QtWidgets.QAction()
        self.action_view_threads.setCheckable(True)
        self.action_view_threads.setChecked(False)
        self.action_view_threads.setText('Show threads')
        self.action_view_threads.setToolTip('Show threads dockwidget')
        self.action_view_default = QtWidgets.QAction()
        self.action_view_default.setText('Restore default')
        self.action_view_default.setToolTip('Restore default view')
        # Dialog actions
        self.action_settings = QtWidgets.QAction()
        self.action_settings.setIcon(QtGui.QIcon(os.path.join(icon_path, 'configure')))
        self.action_settings.setText('Settings')
        self.action_settings.setToolTip('Open settings dialog')
        self.action_about_qudi = QtWidgets.QAction()
        self.action_about_qudi.setIcon(
            QtGui.QIcon(os.path.join(icon_path, 'go-home')))
        self.action_about_qudi.setText('About qudi')
        self.action_about_qudi.setToolTip('Read up about qudi')
        self.action_about_qt = QtWidgets.QAction()
        self.action_about_qt.setIcon(
            QtGui.QIcon(os.path.join(icon_path, 'go-home')))
        self.action_about_qt.setText('About Qt')
        self.action_about_qt.setToolTip('Read up about Qt')

        # Create toolbar
        self.toolbar = QtWidgets.QToolBar()
        self.toolbar.setObjectName('QudiMainWindow Toolbar')
        self.toolbar.setOrientation(QtCore.Qt.Horizontal)
        self.toolbar.addAction(self.action_open_configuration_editor)
        self.toolbar.addAction(self.action_load_configuration)
        self.toolbar.addAction(self.action_reload_qudi)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.action_load_all_modules)
        self.addToolBar(self.toolbar)

        # Create menu bar
        self.menubar = QtWidgets.QMenuBar()
        menu = QtWidgets.QMenu('File')
        menu.addAction(self.action_open_configuration_editor)
        menu.addSeparator()
        menu.addAction(self.action_load_configuration)
        menu.addAction(self.action_reload_qudi)
        menu.addSeparator()
        menu.addAction(self.action_load_all_modules)
        menu.addSeparator()
        menu.addAction(self.action_settings)
        menu.addSeparator()
        menu.addAction(self.action_quit)
        self.menubar.addMenu(menu)
        menu = QtWidgets.QMenu('View')
        menu.addAction(self.action_view_console)
        menu.addAction(self.action_view_log)
        menu.addAction(self.action_view_config)
        menu.addAction(self.action_view_remote)
        menu.addAction(self.action_view_threads)
        menu.addSeparator()
        menu.addAction(self.action_view_default)
        self.menubar.addMenu(menu)
        menu = QtWidgets.QMenu('About')
        menu.addAction(self.action_about_qudi)
        menu.addAction(self.action_about_qt)
        self.menubar.addMenu(menu)
        self.setMenuBar(self.menubar)

        # Create status bar
        self.statusbar = QtWidgets.QStatusBar()
        self.setStatusBar(self.statusbar)
        self.version_label = QtWidgets.QLabel()
        self.version_label.setOpenExternalLinks(True)
        self.statusbar.addWidget(self.version_label)

        # Create dialogues
        self.about_qudi_dialog = AboutQudiDialog()
        self.about_qudi_dialog.setWindowTitle('About qudi')
        self.settings_dialog = SettingsDialog()

        # Create dockwidgets
        self.config_widget = ConfigQTreeWidget()
        self.config_dockwidget = AdvancedDockWidget('Configuration')
        self.config_dockwidget.setWidget(self.config_widget)
        self.config_dockwidget.setAllowedAreas(
            QtCore.Qt.BottomDockWidgetArea | QtCore.Qt.LeftDockWidgetArea
        )
        self.log_widget = LogWidget(debug_mode=debug_mode)
        self.log_dockwidget = AdvancedDockWidget('Log')
        self.log_dockwidget.setWidget(self.log_widget)
        self.log_dockwidget.setAllowedAreas(QtCore.Qt.BottomDockWidgetArea)
        self.remote_widget = RemoteWidget()
        self.remote_dockwidget = AdvancedDockWidget('Remote modules')
        self.remote_dockwidget.setWidget(self.remote_widget)
        self.remote_dockwidget.setAllowedAreas(QtCore.Qt.BottomDockWidgetArea)
        self.threads_widget = QtWidgets.QListView()
        self.threads_dockwidget = AdvancedDockWidget('Threads')
        self.threads_dockwidget.setWidget(self.threads_widget)
        self.threads_dockwidget.setAllowedAreas(QtCore.Qt.BottomDockWidgetArea)
        self.console_widget = RichJupyterWidget()
        self.console_dockwidget = AdvancedDockWidget('Console')
        self.console_dockwidget.setWidget(self.console_widget)
        self.console_dockwidget.setAllowedAreas(
            QtCore.Qt.RightDockWidgetArea | QtCore.Qt.LeftDockWidgetArea)

        # Add dockwidgets to main window
        self.addDockWidget(QtCore.Qt.BottomDockWidgetArea, self.config_dockwidget)
        self.addDockWidget(QtCore.Qt.BottomDockWidgetArea, self.log_dockwidget)
        self.addDockWidget(QtCore.Qt.BottomDockWidgetArea, self.remote_dockwidget)
        self.addDockWidget(QtCore.Qt.BottomDockWidgetArea, self.threads_dockwidget)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.console_dockwidget)

        # Synchronize dockwidget visibility change signals
        self.config_dockwidget.sigClosed.connect(lambda: self.action_view_config.setChecked(False))
        self.log_dockwidget.sigClosed.connect(
            lambda: self.action_view_log.setChecked(False))
        self.remote_dockwidget.sigClosed.connect(
            lambda: self.action_view_remote.setChecked(False))
        self.threads_dockwidget.sigClosed.connect(
            lambda: self.action_view_threads.setChecked(False))
        self.console_dockwidget.sigClosed.connect(
            lambda: self.action_view_console.setChecked(False))
        self.action_view_config.toggled.connect(self.config_dockwidget.setVisible)
        self.action_view_log.toggled.connect(self.log_dockwidget.setVisible)
        self.action_view_remote.toggled.connect(self.remote_dockwidget.setVisible)
        self.action_view_threads.toggled.connect(self.threads_dockwidget.setVisible)
        self.action_view_console.toggled.connect(self.console_dockwidget.setVisible)

        # Connect dialog open signals
        self.action_about_qudi.triggered.connect(self.about_qudi_dialog.open)
        self.action_about_qt.triggered.connect(QtWidgets.QApplication.aboutQt)
        self.action_settings.triggered.connect(self.settings_dialog.exec_)  # modal
        return



# -*- coding: utf-8 -*-
"""
This file contains qudi methods for mathematical operations/transformations.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ('compute_ft', 'ft_windows')

import numpy as np
from scipy.signal import windows as window_func

# Available windows to be applied on signal data before FT.
# To find out the amplitude normalization factor check either the scipy implementation on
#     https://github.com/scipy/scipy/blob/v0.15.1/scipy/signal/windows.py#L336
# or just perform a sum of the window (oscillating parts of the window should be averaged out and
# constant offset factor will remain):
#     MM=1000000  # choose a big number
#     print(sum(signal.hanning(MM))/MM)
ft_windows = {'none': {'func': np.ones, 'ampl_norm': 1.0},
              'hamming': {'func': window_func.hamming, 'ampl_norm': 1.0/0.54},
              'hann': {'func': window_func.hann, 'ampl_norm': 1.0/0.5},
              'blackman': {'func': window_func.blackman, 'ampl_norm': 1.0/0.42},
              'triang': {'func': window_func.triang, 'ampl_norm': 1.0/0.5},
              'flattop': {'func': window_func.flattop, 'ampl_norm': 1.0/0.2156},
              'bartlett': {'func': window_func.bartlett, 'ampl_norm': 1.0/0.5},
              'parzen': {'func': window_func.parzen, 'ampl_norm': 1.0/0.375},
              'bohman': {'func': window_func.bohman, 'ampl_norm': 1.0/0.4052847},
              'blackmanharris': {'func': window_func.blackmanharris, 'ampl_norm': 1.0/0.35875},
              'nuttall': {'func': window_func.nuttall, 'ampl_norm': 1.0/0.3635819},
              'barthann': {'func': window_func.barthann, 'ampl_norm': 1.0/0.5}
              }


def compute_ft(x_val, y_val, zeropad_num=0, window="none", base_corr=True, psd=False):
    """
    Compute the Discrete Fourier Transform (DFT) or Power Spectral Density (PSD) of the input data.

    Parameters
    ----------
    x_val : numpy.array
        1D array representing the x values.
    y_val : numpy.array
        1D array of the same size as x_val, representing the y values.
    zeropad_num : int, optional
        Zeropadding parameter. Number of zeros to add to the end of y_val before performing the Fourier Transform.
        The resulting arrays will have lengths (len(y_val)/2)*(zeropad_num+1). Default is 0.
    window : str, optional
        Window function to apply to y_val before calculating the Fourier Transform.
        Example: 'hann' for Hann window. Default is None (no window applied).
    base_corr : bool, optional
        Flag indicating whether baseline correction should be performed before calculating the Fourier Transform.
        Default is False.
    psd : bool, optional
        Flag indicating whether to compute the Power Spectral Density (PSD) instead of the DFT.
        Default is False (compute DFT).

    Returns
    -------
    tuple(numpy.array, numpy.array)
        Tuple of arrays (dft_x, dft_y) where:
        - dft_x : numpy.array
            Frequencies corresponding to the Fourier Transform.
        - dft_y : numpy.array
            DFT or PSD values. Length of dft_x and dft_y is (len(y_val)/2)*(zeropad_num+1).

    Notes
    -----
    Zeropadding interpolates between DFT/PSD values (Sinc interpolation method) but does not add more information.
    Window functions like Hann are recommended to reduce spectral leakage.
    The sample rate is related to the Fourier transform space as: T = delta_t * N_samples,
    where delta_t is the time interval between points and N_samples is the number of points in the time domain.
    The frequency range of the Fourier Transform is from -f_samplerate/2 to f_samplerate/2.

    PSD describes power distribution over frequency, while DFT shows spectral content (amplitude and phase of harmonics).

    """

    x_val = np.array(x_val)
    y_val = np.array(y_val)

    # Make a baseline correction to avoid a constant offset near zero frequencies. Offset of the
    # y_val from mean corresponds to half the value at fft_y[0]
    corrected_y = y_val
    if base_corr:
        corrected_y = y_val - y_val.mean()

    ampl_norm_fact = 1.0
    # apply window to data to account for spectral leakage:
    if window in ft_windows:
        window_val = ft_windows[window]['func'](len(y_val))
        corrected_y = corrected_y * window_val
        # to get the correct amplitude in the amplitude spectrum
        ampl_norm_fact = ft_windows[window]['ampl_norm']

    # zeropad for sinc interpolation:
    zeropad_arr = np.zeros(len(corrected_y)*(zeropad_num+1))
    zeropad_arr[:len(corrected_y)] = corrected_y

    # Get the amplitude values from the fourier transformed y values.
    fft_y = np.abs(np.fft.fft(zeropad_arr))

    # Power spectral density (PSD) or just amplitude spectrum of fourier signal:
    power_value = 1.0
    if psd:
        power_value = 2.0

    # The factor 2 accounts for the fact that just the half of the spectrum was taken. The
    # ampl_norm_fact is the normalization factor due to the applied window function (the offset
    # value in the window function):
    fft_y = ((2/len(y_val)) * fft_y * ampl_norm_fact)**power_value

    # Due to the sampling theorem you can only identify frequencies at half of the sample rate,
    # therefore the FT contains an almost symmetric spectrum (the asymmetry results from aliasing
    # effects). Therefore take the half of the values for the display.
    middle = int((len(zeropad_arr)+1)//2)

    # sample spacing of x_axis, if x is a time axis than it corresponds to a timestep:
    x_spacing = np.round(x_val[-1] - x_val[-2], 12)

    # use the helper function of numpy to calculate the x_values for the fourier space. That
    # function will handle an occuring devision by 0:
    fft_x = np.fft.fftfreq(len(zeropad_arr), d=x_spacing)
    return abs(fft_x[:middle]), fft_y[:middle]

def normalize(arr:np.ndarray, axis=-1, order=2)->np.ndarray:
    """
    Taken from stack overflow
    https://stackoverflow.com/questions/21030391/how-to-normalize-a-numpy-array-to-a-unit-vector
    """
    l2 = np.atleast_1d(np.linalg.norm(arr, order, axis))
    l2[l2==0] = 1
    return arr / np.expand_dims(l2, axis)




# -*- coding: utf-8 -*-

"""
This file contains a wrapper to display the SpinBox in scientific way

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['ScienDSpinBox', 'ScienSpinBox']

from PySide2 import QtCore, QtGui, QtWidgets
import numpy as np
import re
from decimal import Decimal as D  # Use decimal to avoid accumulating floating-point errors
from decimal import ROUND_FLOOR
import math


class FloatValidator(QtGui.QValidator):
    """
    This is a validator for float values represented as strings in scientific notation.
    (i.e. "1.35e-9", ".24E+8", "14e3" etc.)
    Also supports SI unit prefix like 'M', 'n' etc.
    """

    float_re = re.compile(
        r'(\s*([+-]?)(\d+\.\d+|\.\d+|\d+\.?)([eE][+-]?\d+)?\s?([YZEPTGMkmµunpfazy]?)\s*)'
    )
    group_map = {'match': 0,
                 'sign': 1,
                 'mantissa': 2,
                 'exponent': 3,
                 'si': 4}

    def validate(self, string, position):
        """
        This is the actual validator. It checks whether the current user input is a valid string
        every time the user types a character. There are 3 states that are possible.

        1) Invalid: The current input string is invalid. The user input will not accept the last
           typed character.

        2) Acceptable: The user input is conforming to the regular expression and will be accepted.

        3) Intermediate: The user input is not a valid string yet but on the right track. Use this
           return value to allow the user to type additional characters needed to complete an
           expression (e.g., the decimal point of a float value).

        Parameters
        ----------
        string : str
            The current input string (from a QLineEdit, for example).
        position : int
            The current position of the text cursor.

        Returns
        -------
        QValidator.State
            The returned validator state: QValidator.Invalid, QValidator.Acceptable, or QValidator.Intermediate.
        str
            The input string after validation.
        int
            The updated cursor position after validation.

        """
        # Return intermediate status when empty string is passed or when incomplete "[+-]inf"
        if string.strip() in '+.-.' or string.strip() in list('YZEPTGMkmµunpfazy') or re.match(
                r'[+-]?(in$|i$)', string, re.IGNORECASE):
            return self.Intermediate, string, position

        # Accept input of [+-]inf. Not case sensitive.
        if re.match(r'[+-]?\binf$', string, re.IGNORECASE):
            return self.Acceptable, string.lower(), position

        group_dict = self.get_group_dict(string)
        if group_dict:
            if group_dict['match'] == string:
                return self.Acceptable, string, position
            if string.count('.') > 1:
                return self.Invalid, group_dict['match'], position
            if position > len(string):
                position = len(string)
            if string[position-1] in 'eE-+' and 'i' not in string.lower():
                return self.Intermediate, string, position
            return self.Invalid, group_dict['match'], position
        else:
            if string[position-1] in 'eE-+.' and 'i' not in string.lower():
                return self.Intermediate, string, position
            return self.Invalid, '', position

    def get_group_dict(self, string):
        """
        This method will match the input string with the regular expression of this validator.
        The match groups will be put into a dictionary with string descriptors as keys describing
        the role of the specific group (i.e. mantissa, exponent, si-prefix etc.).

        Parameters
        ----------
        string : str
            Input string to be matched.

        Returns
        -------
        dict
            Dictionary containing groups as items and descriptors as keys (see: self.group_map).

        """
        match = self.float_re.search(string)
        if not match:
            return False
        groups = match.groups()
        group_dict = dict()
        for group_key in self.group_map:
            group_dict[group_key] = groups[self.group_map[group_key]]
        return group_dict

    def fixup(self, text):
        match = self.float_re.search(text)
        if match:
            return match.groups()[0].strip()
        else:
            return ''


class IntegerValidator(QtGui.QValidator):
    """
    This is a validator for int values represented as strings in scientific notation.
    Using engeneering notation only positive exponents are allowed
    (i.e. "1e9", "2E+8", "14e+3" etc.)
    Also supports non-fractional SI unit prefix like 'M', 'k' etc.
    """

    int_re = re.compile(r'(([+-]?\d+)([eE]\+?\d+)?\s?([YZEPTGMk])?\s*)')
    group_map = {'match': 0,
                 'mantissa': 1,
                 'exponent': 2,
                 'si': 3
                 }

    def validate(self, string, position):
        """
        This is the actual validator. It checks whether the current user input is a valid string
        every time the user types a character. There are 3 states that are possible:
        1) Invalid: The current input string is invalid. The user input will not accept the last
                    typed character.
        2) Acceptable: The user input is conforming to the regular expression and will be accepted.
        3) Intermediate: The user input is not a valid string yet but is on the right track. Use this
                         return value to allow the user to type the necessary fill-characters needed
                         to complete an expression (e.g., the decimal point of a float value).

        Parameters
        ----------
        string : str
            The current input string (from a QLineEdit, for example).
        position : int
            The current position of the text cursor.

        Returns
        -------
        QValidator.State or enum
            The returned validator state indicating the validity of the input.
        str
            The input string after validation.
        int
            The cursor position after validation.

        """
        # Return intermediate status when empty string is passed or cursor is at index 0
        if not string.strip() or string.strip() in list('YZEPTGMk'):
            return self.Intermediate, string, position

        group_dict = self.get_group_dict(string)
        if group_dict:
            if group_dict['match'] == string:
                return self.Acceptable, string, position

            if position > len(string):
                position = len(string)
            if string[position-1] in 'eE-+':
                return self.Intermediate, string, position

            return self.Invalid, group_dict['match'], position
        else:
            return self.Invalid, '', position

    def get_group_dict(self, string):
        """
        This method matches the input string with the regular expression defined by this validator.
        It captures match groups into a dictionary with string descriptors as keys describing the role
        of each specific group (e.g., mantissa, exponent, SI-prefix).

        Parameters
        ----------
        string : str
            The input string to be matched against the validator's regular expression.

        Returns
        -------
        dict
            A dictionary containing matched groups as values and descriptors as keys. See `self.group_map`
            for details on the descriptors used.
        """
        match = self.int_re.search(string)
        if not match:
            return False
        groups = match.groups()
        group_dict = dict()
        for group_key in self.group_map:
            group_dict[group_key] = groups[self.group_map[group_key]]
        return group_dict

    def fixup(self, text):
        match = self.int_re.search(text)
        if match:
            return match.groups()[0].strip()
        else:
            return ''


class ScienDSpinBox(QtWidgets.QAbstractSpinBox):
    """
    Wrapper Class from PyQt5 (or QtPy) to display a QDoubleSpinBox in Scientific way.
    Fully supports prefix and suffix functionality of the QDoubleSpinBox.
    Has built-in functionality to invoke the displayed number precision from the user input.

    This class can be directly used in Qt Designer by promoting the QDoubleSpinBox to ScienDSpinBox.
    State the path to this file (in python style, i.e. dots are separating the directories) as the
    header file and use the name of the present class.
    """

    valueChanged = QtCore.Signal(object)

    # The maximum number of decimals to allow. Be careful when changing this number since
    # the decimal package has by default a limited accuracy.
    __max_decimals = 20
    # Dictionary mapping the si-prefix to a scaling factor as decimal.Decimal (exact value)
    _unit_prefix_dict = {
        'y': D('1e-24'),
        'z': D('1e-21'),
        'a': D('1e-18'),
        'f': D('1e-15'),
        'p': D('1e-12'),
        'n': D('1e-9'),
        'µ': D('1e-6'),
        'm': D('1e-3'),
        '': D('1'),
        'k': D('1e3'),
        'M': D('1e6'),
        'G': D('1e9'),
        'T': D('1e12'),
        'P': D('1e15'),
        'E': D('1e18'),
        'Z': D('1e21'),
        'Y': D('1e24')
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__value = D(0)
        self.__minimum = -np.inf
        self.__maximum = np.inf
        self.__decimals = 2  # default in QtDesigner
        self.__prefix = ''
        self.__suffix = ''
        self.__singleStep = D('0.1')  # must be precise Decimal always, no conversion from float
        self.__minimalStep = D(0)  # must be precise Decimal always, no conversion from float
        self.__cached_value = None  # a temporary variable for restore functionality
        self._dynamic_stepping = True
        self._dynamic_precision = True
        self._assumed_unit_prefix = None  # To assume one prefix. This is only used if no prefix would be out of range
        self._is_valid = True  # A flag property to check if the current value is valid.
        self.disable_wheel = False
        self.validator = FloatValidator()
        self.lineEdit().textEdited.connect(self.update_value)
        self.update_display()

    @property
    def dynamic_stepping(self):
        """
        Flag indicating whether dynamic (logarithmic) stepping should be used or fixed steps.

        Returns
        -------
        bool
            True if dynamic stepping is enabled, False otherwise.
        """
        return bool(self._dynamic_stepping)

    @dynamic_stepping.setter
    def dynamic_stepping(self, use_dynamic_stepping):
        """
        Flag indicating whether dynamic (logarithmic) stepping should be used or fixed steps.

        Parameters
        ----------
        use_dynamic_stepping : bool
            True to use dynamic stepping (logarithmic), False to use constant steps.

        Returns
        -------
        bool
            True if dynamic stepping is enabled, False otherwise.
        """
        use_dynamic_stepping = bool(use_dynamic_stepping)
        self._dynamic_stepping = use_dynamic_stepping

    @property
    def dynamic_precision(self):
        """
        Flag indicating whether dynamic decimal precision should be used based on user input.

        Returns
        -------
        bool
            True to use dynamic decimal precision, False to use fixed precision.
        """
        return bool(self._dynamic_precision)

    @dynamic_precision.setter
    def dynamic_precision(self, use_dynamic_precision):
        """
        Flag indicating whether dynamic decimal precision should be used based on user input.

        Parameters
        ----------
        use_dynamic_precision : bool
            True to use dynamic precision (invoked from user input), False to use fixed precision.

        Returns
        -------
        bool
            True to use dynamic precision, False to use fixed precision.
        """
        use_dynamic_precision = bool(use_dynamic_precision)
        self._dynamic_precision = use_dynamic_precision

    @property
    def assumed_unit_prefix(self):
        """
        Default unit prefix for text input.

        Returns
        -------
        str or None
            The default unit prefix string if set, otherwise None.
        """
        return self._assumed_unit_prefix

    @assumed_unit_prefix.setter
    def assumed_unit_prefix(self, unit_prefix):
        """
        Default unit prefix used for text input.

        Parameters
        ----------
        unit_prefix : str or None
            The unit prefix to set as default for text input. If None, the default prefix is cleared.
        """
        if unit_prefix is None or unit_prefix in self._unit_prefix_dict:
            self._assumed_unit_prefix = unit_prefix
        if unit_prefix == 'u':  # in case of encoding problems
            self._assumed_unit_prefix = 'µ'

    @property
    def is_valid(self):
        """
        Flag indicating if the currently available value is valid.

        Returns
        -------
        bool
            True if the current value is valid, False otherwise. Returns False if there has been an attempt
            to set NaN as the current value; True after a valid value has been set.
        """
        return bool(self._is_valid)

    def update_value(self):
        """
        This method will grab the currently shown text from the QLineEdit and interpret it.
        Range checking is performed on the value afterwards.
        If a valid value can be derived, it will set this value as the current value
        (if it has changed) and emit the valueChanged signal.
        Note that the comparison between old and new value is done by comparing the float
        representations of both values and not by comparing them as Decimals.
        The valueChanged signal will only emit if the actual float representation has changed since
        Decimals are only internally used and the rest of the program won't notice a slight change
        in the Decimal that can't be resolved in a float.
        In addition it will cache the old value provided the cache is empty to be able to restore
        it later on.
        """
        text = self.cleanText()
        value = self.valueFromText(text)
        if value is False:
            return
        value, in_range = self.check_range(value)
        # if the value is out of range, then only use assumed unit prefix
        if not in_range and self._assumed_unit_prefix is not None:
            value = self.valueFromText(text, use_assumed_unit_prefix=True)
            value, in_range = self.check_range(value)
        # save old value to be able to restore it later on
        if self.__cached_value is None:
            self.__cached_value = self.__value

        if float(value) != self.value():
            self.__value = value
            self.valueChanged.emit(self.value())
        else:
            self.__value = value
        self._is_valid = True

    def value(self):
        """
        Getter method to obtain the current value as float.

        Returns
        -------
        float
            The current value of the SpinBox as a float.
        """
        return float(self.__value)

    def setValue(self, value):
        """
        Setter method to programmatically set the current value. For best robustness pass the value
        as string or Decimal in order to be lossless cast into Decimal.
        Will perform range checking and ignore NaN values.
        Will emit valueChanged if the new value is different from the old one.
        When using dynamic decimals precision, this method will also try to invoke the optimal
        display precision by checking for a change in the displayed text.
        """
        try:
            value = D(value)
        except TypeError:
            if 'int' in type(value).__name__:
                value = int(value)
            elif 'float' in type(value).__name__:
                value = float(value)
            else:
                raise
            value = D(value)

        # catch NaN values and set the "is_valid" flag to False until a valid value is set again.
        if value.is_nan():
            self._is_valid = False
            return

        value, in_range = self.check_range(value)

        if self.__value != value or not self.is_valid:
            # Try to increase decimals when the value has changed but no change in display detected.
            # This will only be executed when the dynamic precision flag is set
            if self.value() != float(value) and self.dynamic_precision and not value.is_infinite():
                old_text = self.cleanText()
                new_text = self.textFromValue(value).strip()
                current_dec = self.decimals()
                while old_text == new_text:
                    if self.__decimals > self.__max_decimals:
                        self.__decimals = current_dec
                        break
                    self.__decimals += 1
                    new_text = self.textFromValue(value).strip()
            self.__value = value
            self._is_valid = True
            self.update_display()
            self.valueChanged.emit(self.value())

    def setProperty(self, prop, val):
        """
        For compatibility with QtDesigner. Initializes the value through this method.

        Parameters
        ----------
        prop : type
            Description of the parameter 'prop'.
        val : type
            Description of the parameter 'val'.
        """
        if prop == 'value':
            self.setValue(val)
        else:
            raise UserWarning('setProperty in scientific spinboxes only works for "value".')

    def check_range(self, value):
        """
        Helper method to check if the passed value is within the set minimum and maximum value bounds.

        If outside of bounds, the returned value will be clipped to the nearest boundary.

        Parameters
        ----------
        value : float or Decimal
            Number to be checked.

        Returns
        -------
        Decimal
            The corrected value.
        bool
            Flag indicating if the value has been changed (`True`) or not (`False`).
        """

        if value < self.__minimum:
            new_value = self.__minimum
            in_range = False
        elif value > self.__maximum:
            new_value = self.__maximum
            in_range = False
        else:
            in_range = True
        if not in_range:
            value = D(new_value)
        return value, in_range

    def minimum(self):
        return float(self.__minimum)

    def setMinimum(self, minimum):
        """
        Setter method to set the minimum value allowed in the SpinBox.

        Parameters
        ----------
        minimum : float
            The minimum value to be set. Input will be converted to float before being stored.
        """

        # Ignore NaN values
        if self._check_nan(float(minimum)):
            return

        self.__minimum = float(minimum)
        if self.__minimum > self.value():
            self.setValue(self.__minimum)

    def maximum(self):
        return float(self.__maximum)

    def setMaximum(self, maximum):
        """
        Setter method to set the maximum value allowed in the SpinBox.

        Parameters
        ----------
        maximum : float
            The maximum value to be set. Input will be converted to float before being stored.
        """

        # Ignore NaN values
        if self._check_nan(float(maximum)):
            return

        self.__maximum = float(maximum)
        if self.__maximum < self.value():
            self.setValue(self.__maximum)

    def setRange(self, minimum, maximum):
        """
        Convenience method for compliance with Qt SpinBoxes.
        Essentially a wrapper to call both self.setMinimum and self.setMaximum.

        Parameters
        ----------
        minimum : float
            The minimum value to be set.
        maximum : float
            The maximum value to be set.
        """

        self.setMinimum(minimum)
        self.setMaximum(maximum)

    def decimals(self):
        return self.__decimals

    def setDecimals(self, decimals, dynamic_precision=True):
        """
        Set the number of displayed digits after the decimal point and specify dynamic precision.

        Parameters
        ----------
        decimals : int
            The number of decimals to be displayed.
        dynamic_precision : bool
            Flag indicating whether dynamic precision functionality should be used:
            - If True, the number of decimals will be determined dynamically from user input until
              explicitly set by calling this method or entering user text.
            - If False, the specified number of decimals will be fixed and will not change automatically.

        Returns
        -------
        None
        """
        decimals = int(decimals)
        # Restrict the number of fractional digits to a maximum of self.__max_decimals = 20.
        # Beyond that the number is not very meaningful anyways due to machine precision.
        if decimals < 0:
            decimals = 0
        elif decimals > self.__max_decimals:
            decimals = self.__max_decimals
        self.__decimals = decimals
        # Set the flag for using dynamic precision (decimals invoked from user input)
        self.dynamic_precision = dynamic_precision

    def prefix(self):
        return self.__prefix

    def setPrefix(self, prefix):
        """
        Set a string to be shown as non-editable prefix in the spinbox.

        Parameters
        ----------
        prefix : str
            The prefix string to be displayed.

        Returns
        -------
        None
        """

        self.__prefix = str(prefix)
        self.update_display()

    def suffix(self):
        return self.__suffix

    def setSuffix(self, suffix):
        """
        Set a string to be shown as non-editable suffix in the spinbox.
        This suffix will come right after the si-prefix.

        Parameters
        ----------
        suffix : str
            The suffix string to be displayed after the si-prefix.

        Returns
        -------
        None
        """

        self.__suffix = str(suffix)
        self.update_display()

    def singleStep(self):
        return float(self.__singleStep)

    def setSingleStep(self, step, dynamic_stepping=True):
        """
        Set the stepping behavior of the spinbox (e.g., when using the mouse wheel).

        When dynamic_stepping=True, the spinbox will perform logarithmic steps according to the
        current order of magnitude of the values. The step parameter then specifies the step size
        relative to the value's order of magnitude. For example, step=0.1 would increment the second
        most significant digit by one.

        When dynamic_stepping=False, the step parameter specifies an absolute step size. This means
        that each time a step is performed, this value is added or subtracted from the current value.

        For maximum robustness and consistency, it is strongly recommended to pass step as a Decimal
        or string to ensure lossless conversion to Decimal.

        Parameters
        ----------
        step : Decimal or str
            The (relative) step size to set. For dynamic_stepping=True, this is relative to the order
            of magnitude of the current value. For dynamic_stepping=False, this is an absolute step size.
        dynamic_stepping : bool
            Flag indicating the use of dynamic stepping (True) or constant stepping (False).

        Returns
        -------
        None
        """
        try:
            step = D(step)
        except TypeError:
            if 'int' in type(step).__name__:
                step = int(step)
            elif 'float' in type(step).__name__:
                step = float(step)
            else:
                raise
            step = D(step)

        # ignore NaN and infinity values
        if not step.is_nan() and not step.is_infinite():
            self.__singleStep = step

        self.dynamic_stepping = dynamic_stepping

    def minimalStep(self):
        return float(self.__minimalStep)

    def setMinimalStep(self, step):
        """
        Method used to set a minimal step size.

        When the absolute step size has been calculated in either dynamic or constant step mode,
        this value is checked against the minimal step size. If it is smaller then the minimal step
        size is chosen over the calculated step size. This ensures that no step taken can be
        smaller than minimalStep.

        Set this value to 0 for no minimal step size.

        For maximum roboustness and consistency it is strongly recommended to pass step as Decimal
        or string in order to be converted lossless to Decimal.

        Parameters
        ----------
        step : Decimal|str
            The minimal step size to be set.

        Returns
        -------
        None
        """

        try:
            step = D(step)
        except TypeError:
            if 'int' in type(step).__name__:
                step = int(step)
            elif 'float' in type(step).__name__:
                step = float(step)
            else:
                raise
            step = D(step)

        # ignore NaN and infinity values
        if not step.is_nan() and not step.is_infinite():
            self.__minimalStep = step

    def cleanText(self):
        """
        Compliance method from Qt SpinBoxes.
        Returns the currently shown text from the QLineEdit without prefix and suffix and stripped
        from leading or trailing whitespaces.

        Returns
        -------
        str
            Currently shown text stripped from suffix and prefix.
        """

        text = self.text().strip()
        if self.__prefix and text.startswith(self.__prefix):
            text = text[len(self.__prefix):]
        if self.__suffix and text.endswith(self.__suffix):
            text = text[:-len(self.__suffix)]
        return text.strip()

    def update_display(self):
        """
        This helper method updates the shown text based on the current value.
        Because this method is only called upon finishing an editing procedure, the eventually
        cached value gets deleted.
        """
        text = self.textFromValue(self.value())
        text = self.__prefix + text + self.__suffix
        self.lineEdit().setText(text)
        self.__cached_value = None  # clear cached value
        self.lineEdit().setCursorPosition(0)  # Display the most significant part of the number

    def keyPressEvent(self, event):
        """
        This method catches all keyboard press events triggered by the user.

        Can be used to alter the behaviour of certain key events from the default
        implementation of QAbstractSpinBox.

        Parameters
        ----------
        event : QKeyEvent
            A Qt QKeyEvent instance holding the event information
        """

        # Restore cached value upon pressing escape and lose focus.
        if event.key() == QtCore.Qt.Key_Escape:
            if self.__cached_value is not None:
                self.__value = self.__cached_value
                self.valueChanged.emit(self.value())
            self.clearFocus()  # This will also trigger editingFinished
            return

        # Update display upon pressing enter/return before processing the event in the default way.
        if event.key() == QtCore.Qt.Key_Enter or event.key() == QtCore.Qt.Key_Return:
            self.update_display()

        if (QtCore.Qt.ControlModifier | QtCore.Qt.MetaModifier) & event.modifiers():
            super().keyPressEvent(event)
            return

        # The rest is to avoid editing suffix and prefix
        if len(event.text()) > 0:
            # Allow editing of the number or SI-prefix even if part of the prefix/suffix is selected.
            if self.lineEdit().selectedText():
                sel_start = self.lineEdit().selectionStart()
                sel_end = sel_start + len(self.lineEdit().selectedText())
                min_start = len(self.__prefix)
                max_end = len(self.__prefix) + len(self.cleanText())
                if sel_start < min_start:
                    sel_start = min_start
                if sel_end > max_end:
                    sel_end = max_end
                self.lineEdit().setSelection(sel_start, sel_end - sel_start)
            else:
                cursor_pos = self.lineEdit().cursorPosition()
                begin = len(self.__prefix)
                end = len(self.text()) - len(self.__suffix)
                if cursor_pos < begin:
                    self.lineEdit().setCursorPosition(begin)
                elif cursor_pos > end:
                    self.lineEdit().setCursorPosition(end)

        if event.key() == QtCore.Qt.Key_Left:
            if self.lineEdit().cursorPosition() == len(self.__prefix):
                return
        if event.key() == QtCore.Qt.Key_Right:
            if self.lineEdit().cursorPosition() == len(self.text()) - len(self.__suffix):
                return
        if event.key() == QtCore.Qt.Key_Home:
            self.lineEdit().setCursorPosition(len(self.__prefix))
            return
        if event.key() == QtCore.Qt.Key_End:
            self.lineEdit().setCursorPosition(len(self.text()) - len(self.__suffix))
            return

        super().keyPressEvent(event)

    def focusInEvent(self, event):
        super().focusInEvent(event)
        self.selectAll()
        return

    def focusOutEvent(self, event):
        self.update_display()
        super().focusOutEvent(event)
        return

    def paintEvent(self, ev):
        """
        Add drawing of a red frame around the spinbox if the is_valid flag is False
        """
        super().paintEvent(ev)

        # draw red frame if is_valid = False
        if not self.is_valid:
            pen = QtGui.QPen()
            pen.setColor(QtGui.QColor(200, 50, 50))
            pen.setWidth(2)

            p = QtGui.QPainter(self)
            p.setRenderHint(p.Antialiasing)
            p.setPen(pen)
            p.drawRoundedRect(self.rect().adjusted(2, 2, -2, -2), 4, 4)
            p.end()

    def validate(self, text, position):
        """
        Access method to the validator. See FloatValidator class for more information.

        Parameters
        ----------
        text : str
            String to be validated.
        position : int
            Current text cursor position.

        Returns
        -------
        QValidator::State
            The returned validator state.
        str
            The input string.
        int
            The cursor position.
        """

        begin = len(self.__prefix)
        end = len(text) - len(self.__suffix)
        if position < begin:
            position = begin
        elif position > end:
            position = end

        if self.__prefix and text.startswith(self.__prefix):
            text = text[len(self.__prefix):]
        if self.__suffix and text.endswith(self.__suffix):
            text = text[:-len(self.__suffix)]

        state, string, position = self.validator.validate(text, position)

        text = self.__prefix + string + self.__suffix

        end = len(text) - len(self.__suffix)
        if position > end:
            position = end

        return state, text, position

    def fixup(self, text):
        """
        Takes an invalid string and tries to fix it in order to pass validation.
        The returned string is not guaranteed to pass validation.

        Parameters
        ----------
        text : str
            A string that has not passed validation and needs to be fixed.

        Returns
        -------
        str
            The resulting string from the fix attempt.
        """

        return self.validator.fixup(text)

    def valueFromText(self, text, use_assumed_unit_prefix=False):
        """
        Convert a string displayed in the SpinBox into a Decimal value.

        The input string is already stripped of prefix and suffix.
        Only the si-prefix may be present.

        Parameters
        ----------
        text : str
            The display string to be converted into a numeric value.
            This string must conform to the validator.

        Returns
        -------
        Decimal
            The numeric value converted from the input string.
        """
        # Check for infinite value
        if 'inf' in text.lower():
            if text.startswith('-'):
                return D('-inf')
            else:
                return D('inf')

        # Handle "normal" (non-infinite) input
        group_dict = self.validator.get_group_dict(text)
        if not group_dict:
            return False

        if not group_dict['mantissa']:
            return False

        si_prefix = group_dict['si']
        if si_prefix is None:
            si_prefix = ''
        if si_prefix == '' and use_assumed_unit_prefix and self._assumed_unit_prefix is not None:
            si_prefix = self._assumed_unit_prefix
        si_scale = self._unit_prefix_dict[si_prefix.replace('u', 'µ')]

        if group_dict['sign'] is not None:
            unscaled_value_str = group_dict['sign'] + group_dict['mantissa']
        else:
            unscaled_value_str = group_dict['mantissa']
        if group_dict['exponent'] is not None:
            unscaled_value_str += group_dict['exponent']

        value = D(unscaled_value_str) * si_scale

        # Try to extract the precision the user intends to use
        if self.dynamic_precision:
            split_mantissa = group_dict['mantissa'].split('.')
            if len(split_mantissa) == 2:
                self.setDecimals(max(len(split_mantissa[1]), 1))
            else:
                self.setDecimals(1)  # Minimum number of digits is 1

        return value

    def textFromValue(self, value):
        """
        This method is responsible for the mapping of the underlying value to a string to display
        in the SpinBox. Suffix and Prefix must not be handled here, just the si-Prefix.

        The main problem here is, that a scaled float with a suffix is represented by a different
        machine precision than the total value. This method is so complicated because it represents
        the actual precision of the value as float and not the precision of the scaled si float.
        '{:.20f}'.format(value) shows different digits than
        '{:.20f} {}'.format(scaled_value, si_prefix)

        Parameters
        ----------
        value : float or decimal.Decimal
            The numeric value to be formatted into a string.

        Returns
        -------
        str
            The formatted string representing the input value.
        """
        # Catch infinity value
        if np.isinf(float(value)):
            if value < 0:
                return '-inf '
            else:
                return 'inf '

        sign = '-' if value < 0 else ''
        fractional, integer = math.modf(abs(value))
        integer = int(integer)
        si_prefix = ''
        prefix_index = 0
        if integer != 0:
            integer_str = str(integer)
            fractional_str = ''
            while len(integer_str) > 3:
                fractional_str = integer_str[-3:] + fractional_str
                integer_str = integer_str[:-3]
                if prefix_index < 8:
                    si_prefix = 'kMGTPEZY'[prefix_index]
                else:
                    si_prefix = 'e{0:d}'.format(3 * (prefix_index + 1))
                prefix_index += 1
            # Truncate and round to set number of decimals
            # Add digits from fractional if it's not already enough for set self.__decimals
            if self.__decimals < len(fractional_str):
                round_indicator = int(fractional_str[self.__decimals])
                fractional_str = fractional_str[:self.__decimals]
                if round_indicator >= 5:
                    if not fractional_str:
                        fractional_str = '1'
                    else:
                        fractional_str = str(int(fractional_str) + 1)
            elif self.__decimals == len(fractional_str):
                if fractional >= 0.5:
                    if fractional_str:
                        fractional_int = int(fractional_str) + 1
                        fractional_str = str(fractional_int)
                    else:
                        fractional_str = '1'
            elif self.__decimals > len(fractional_str):
                digits_to_add = self.__decimals - len(fractional_str) # number of digits to add
                fractional_tmp_str = ('{0:.' + str(digits_to_add) + 'f}').format(fractional)
                if fractional_tmp_str.startswith('1'):
                    if fractional_str:
                        fractional_str = str(int(fractional_str) + 1) + '0' * digits_to_add
                    else:
                        fractional_str = '1' + '0' * digits_to_add
                else:
                    fractional_str += fractional_tmp_str.split('.')[1]
            # Check if the rounding has overflown the fractional part into the integer part
            if len(fractional_str) > self.__decimals:
                integer_str = str(int(integer_str) + 1)
                fractional_str = '0' * self.__decimals
        elif fractional == 0.0:
            fractional_str = '0' * self.__decimals
            integer_str = '0'
        else:
            # determine the order of magnitude by comparing the fractional to unit values
            prefix_index = 1
            magnitude = 1e-3
            si_prefix = 'm'
            while magnitude > fractional:
                prefix_index += 1
                magnitude = magnitude ** prefix_index
                if prefix_index <= 8:
                    si_prefix = 'mµnpfazy'[prefix_index - 1]  # use si-prefix if possible
                else:
                    si_prefix = 'e-{0:d}'.format(3 * prefix_index)  # use engineering notation
            # Get the string representation of all needed digits from the fractional part of value.
            digits_needed = 3 * prefix_index + self.__decimals
            helper_str = ('{0:.' + str(digits_needed) + 'f}').format(fractional)
            overflow = bool(int(helper_str.split('.')[0]))
            helper_str = helper_str.split('.')[1]
            if overflow:
                integer_str = '1000'
                fractional_str = '0' * self.__decimals
            elif (prefix_index - 1) > 0 and helper_str[3 * (prefix_index - 1) - 1] != '0':
                integer_str = '1000'
                fractional_str = '0' * self.__decimals
            else:
                integer_str = str(int(helper_str[:3 * prefix_index]))
                fractional_str = helper_str[3 * prefix_index:3 * prefix_index + self.__decimals]

        # Create the actual string representation of value scaled in a scientific way
        space = '' if si_prefix.startswith('e') else ' '
        if self.__decimals > 0:
            string = '{0}{1}.{2}{3}{4}'.format(sign, integer_str, fractional_str, space, si_prefix)
        else:
            string = '{0}{1}{2}{3}'.format(sign, integer_str, space, si_prefix)
        return string

    def stepEnabled(self):
        """
        Enables stepping (mouse wheel, arrow up/down, clicking, PgUp/Down) by default.
        """
        return self.StepUpEnabled | self.StepDownEnabled

    def wheelEvent(self, event):
        """
        Overwriting wheel event, such that with the class variable disable_wheel = True the stepping with the mouse
        wheel is turned off and the wheel event is passed to the parent widget.
        :param event:
        """
        if self.disable_wheel:
            event.ignore()
        else:
            super().wheelEvent(event)

    def stepBy(self, steps):
        """
        This method is responsible for incrementing the value of the SpinBox when the user triggers a step
        (by pressing PgUp/PgDown/Up/Down, MouseWheel movement or clicking on the arrows). It should handle
        the case when the new to-set value is out of bounds. Also, the absolute value of a single step
        increment should be handled here. It is essential to avoid accumulating rounding errors and/or
        discrepancies between self.value and the displayed text.

        Parameters
        ----------
        steps : int
            Number of steps to increment (NOT the absolute step size).
        """

        # Ignore stepping for infinity values
        if self.__value.is_infinite():
            return

        n = D(int(steps))  # n must be integral number of steps.
        s = [D(-1), D(1)][n >= 0]  # determine sign of step
        value = self.__value  # working copy of current value
        if self.dynamic_stepping:
            for i in range(int(abs(n))):
                if value == 0:
                    if self.__minimalStep == 0:
                        if np.isinf(self.__minimum) or np.isinf(self.__maximum):
                            step = D('0.01')
                        else:
                            step = D((self.__maximum - self.__minimum)/10000)
                    else:
                        step = self.__minimalStep
                else:
                    vs = [D(-1), D(1)][value >= 0]
                    fudge = D('1.01') ** (s * vs)  # fudge factor. At some places, the step size
                                                   # depends on the step sign.
                    exp = abs(value * fudge).log10().quantize(1, rounding=ROUND_FLOOR)
                    step = self.__singleStep * D(10) ** exp
                    if self.__minimalStep > 0:
                        step = max(step, self.__minimalStep)
                value += s * step
        else:
            value = value + n * max(self.__minimalStep, self.__singleStep)
        self.setValue(value)

    def selectAll(self):
        begin = len(self.__prefix)
        text = self.cleanText()
        if text.endswith(' '):
            selection_length = len(text) + 1
        elif len(text) > 0 and text[-1] in self._unit_prefix_dict:
            selection_length = len(text) - 1
        else:
            selection_length = len(text)
        self.lineEdit().setSelection(begin, selection_length)

    @staticmethod
    def _check_nan(value):
        """
        Helper method to check if the passed float value is NaN.
        Makes use of the fact that NaN values will always compare to false, even with itself.

        Parameters
        ----------
        value : Decimal or float
            Value to be checked for NaN.

        Returns
        -------
        bool
            True if the value is NaN, False otherwise.
        """

        return not value == value


class ScienSpinBox(QtWidgets.QAbstractSpinBox):
    """
    Wrapper Class from PyQt5 (or QtPy) to display a QSpinBox in Scientific way.
    Fully supports prefix and suffix functionality of the QSpinBox.

    This class can be directly used in Qt Designer by promoting the QSpinBox to ScienSpinBox.
    State the path to this file (in python style, i.e. dots are separating the directories) as the
    header file and use the name of the present class.
    """

    valueChanged = QtCore.Signal(object)
    # Dictionary mapping the si-prefix to a scaling factor as integer (exact value)
    _unit_prefix_dict = {
        '': 1,
        'k': 10 ** 3,
        'M': 10 ** 6,
        'G': 10 ** 9,
        'T': 10 ** 12,
        'P': 10 ** 15,
        'E': 10 ** 18,
        'Z': 10 ** 21,
        'Y': 10 ** 24
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__value = 0
        self.__minimum = -2 ** 63  # Use a 64bit integer size by default.
        self.__maximum = 2 ** 63 - 1  # Use a 64bit integer size by default.
        self.__prefix = ''
        self.__suffix = ''
        self.__singleStep = 1
        self.__minimalStep = 1
        self.__cached_value = None  # a temporary variable for restore functionality
        self._dynamic_stepping = True
        self.disable_wheel = False
        self.validator = IntegerValidator()
        self.lineEdit().textEdited.connect(self.update_value)
        self.update_display()

    @property
    def dynamic_stepping(self):
        """
        Property indicating whether dynamic (logarithmic) stepping is enabled or fixed steps are used.

        Returns
        -------
        bool
            True if dynamic stepping is enabled, False if fixed steps are used.
        """

        return bool(self._dynamic_stepping)

    @dynamic_stepping.setter
    def dynamic_stepping(self, use_dynamic_stepping):
        """
        This property indicates whether dynamic (logarithmic) stepping should be used or fixed steps.

        Parameters
        ----------
        use_dynamic_stepping : bool
            Flag to determine the stepping method:
            - True: Use dynamic (logarithmic) stepping.
            - False: Use fixed steps.
        """

        use_dynamic_stepping = bool(use_dynamic_stepping)
        self._dynamic_stepping = use_dynamic_stepping

    def update_value(self):
        """
        This method will grab the currently shown text from the QLineEdit and interpret it.
        Range checking is performed on the value afterwards.
        If a valid value can be derived, it will set this value as the current value
        (if it has changed) and emit the valueChanged signal.
        In addition it will cache the old value provided the cache is empty to be able to restore
        it later on.
        """
        text = self.cleanText()
        value = self.valueFromText(text)
        if value is False:
            return
        value, in_range = self.check_range(value)

        # save old value to be able to restore it later on
        if self.__cached_value is None:
            self.__cached_value = self.__value

        if value != self.value():
            self.__value = value
            self.valueChanged.emit(self.value())

    def value(self):
        """
        Getter method to obtain the current value as an integer.

        Returns
        -------
        int
            The current value of the SpinBox.
        """
        return int(self.__value)

    def setValue(self, value):
        """
        Setter method to programmatically set the current value.
        Will perform range checking and ignore NaN values.
        Will emit valueChanged if the new value is different from the old one.
        """
        if value is np.nan:
            return

        value = int(value)

        value, in_range = self.check_range(value)

        if self.__value != value:
            self.__value = value
            self.update_display()
            self.valueChanged.emit(self.value())

    def setProperty(self, prop, val):
        """
        For compatibility with QtDesigner. Initializes the value through this method.

        Parameters
        ----------
        prop : str
            Property name.
        val : object
            Value to set for the property.
        """
        if prop == 'value':
            self.setValue(val)
        else:
            raise UserWarning('setProperty in scientific spinboxes only works for "value".')

    def check_range(self, value):
        """
        Helper method to check if the passed value is within the set minimum and maximum value bounds.
        If outside of bounds, the returned value will be clipped to the nearest boundary.

        Parameters
        ----------
        value : int
            Number to be checked.

        Returns
        -------
        (int, bool)
            The corrected value and a flag indicating if the value has been changed:
            - False: Value has been corrected (clipped).
            - True: Value remains unchanged.
        """
        if value < self.__minimum:
            new_value = self.__minimum
            in_range = False
        elif value > self.__maximum:
            new_value = self.__maximum
            in_range = False
        else:
            in_range = True
        if not in_range:
            value = int(new_value)
        return value, in_range

    def minimum(self):
        return int(self.__minimum)

    def setMinimum(self, minimum):
        """
        Setter method to set the minimum value allowed in the SpinBox.
        Input will be converted to int before being stored.

        Parameters
        ----------
        minimum : int
            The minimum value to be set.
        """
        self.__minimum = int(minimum)
        if self.__minimum > self.value():
            self.setValue(self.__minimum)

    def maximum(self):
        return int(self.__maximum)

    def setMaximum(self, maximum):
        """
        Setter method to set the maximum value allowed in the SpinBox.
        Input will be converted to int before being stored.

        Parameters
        ----------
        maximum : int
            The maximum value to be set.
        """
        self.__maximum = int(maximum)
        if self.__maximum < self.value():
            self.setValue(self.__maximum)

    def setRange(self, minimum, maximum):
        """
        Convenience method for compliance with Qt SpinBoxes.
        Essentially a wrapper to call both self.setMinimum and self.setMaximum.

        Parameters
        ----------
        minimum : int
            The minimum value to be set.
        maximum : int
            The maximum value to be set.
        """
        self.setMinimum(minimum)
        self.setMaximum(maximum)

    def prefix(self):
        return self.__prefix

    def setPrefix(self, prefix):
        """
        Set a string to be shown as a non-editable prefix in the spinbox.

        Parameters
        ----------
        prefix : str
            The prefix string to be set.
        """
        self.__prefix = str(prefix)
        self.update_display()

    def suffix(self):
        return self.__suffix

    def setSuffix(self, suffix):
        """
        Set a string to be shown as a non-editable suffix in the spinbox.
        This suffix will come right after the SI-prefix.

        Parameters
        ----------
        suffix : str
            The suffix string to be set.
        """
        self.__suffix = str(suffix)
        self.update_display()

    def singleStep(self):
        return int(self.__singleStep)

    def setSingleStep(self, step, dynamic_stepping=True):
        """
        Method to set the stepping behavior of the spinbox (e.g., when moving the mouse wheel).

        Parameters
        ----------
        step : int
            The absolute step size to set. Ignored if dynamic_stepping=True.
        dynamic_stepping : bool
            Flag indicating the stepping method:
            - True: Use dynamic stepping (logarithmic steps according to current order of magnitude).
            - False: Use constant stepping (step parameter specifies absolute step size).

        Returns
        -------
        None

        Notes
        -----
        When dynamic_stepping=True, the step parameter is ignored. The spinbox will increment the second
        most significant digit by one.
        """
        if step < 1:
            step = 1
        self.__singleStep = int(step)
        self.dynamic_stepping = dynamic_stepping

    def minimalStep(self):
        return int(self.__minimalStep)

    def setMinimalStep(self, step):
        """
        Method used to set a minimal step size.

        Parameters
        ----------
        step : int
            The minimal step size to be set.

        Returns
        -------
        None

        Notes
        -----
        When the absolute step size has been calculated in either dynamic or constant step mode,
        this value is checked against the minimal step size. If it is smaller, then the minimal step
        size is chosen over the calculated step size. This ensures that no step taken can be
        smaller than minimalStep.
        Minimal step size can't be smaller than 1 for integers.
        """
        if step < 1:
            step = 1
        self.__minimalStep = int(step)

    def cleanText(self):
        """
        Compliance method from Qt SpinBoxes.
        Returns the currently shown text from the QLineEdit without prefix and suffix and stripped
        from leading or trailing whitespaces.

        Returns
        -------
        str
            Currently shown text stripped from suffix and prefix.
        """
        text = self.text().strip()
        if self.__prefix and text.startswith(self.__prefix):
            text = text[len(self.__prefix):]
        if self.__suffix and text.endswith(self.__suffix):
            text = text[:-len(self.__suffix)]
        return text.strip()

    def update_display(self):
        """
        This helper method updates the shown text based on the current value.
        Because this method is only called upon finishing an editing procedure, the eventually
        cached value gets deleted.
        """
        text = self.textFromValue(self.value())
        text = self.__prefix + text + self.__suffix
        self.lineEdit().setText(text)
        self.__cached_value = None  # clear cached value
        self.lineEdit().setCursorPosition(0)  # Display the most significant part of the number

    def keyPressEvent(self, event):
        """
        This method catches all keyboard press events triggered by the user. It can be used to alter
        the behavior of certain key events from the default implementation of QAbstractSpinBox.

        Parameters
        ----------
        event : QKeyEvent
            A Qt QKeyEvent instance holding the event information.
        """
        # Restore cached value upon pressing escape and lose focus.
        if event.key() == QtCore.Qt.Key_Escape:
            if self.__cached_value is not None:
                self.__value = self.__cached_value
                self.valueChanged.emit(self.value())
            self.clearFocus()  # This will also trigger editingFinished

        # Update display upon pressing enter/return before processing the event in the default way.
        if event.key() == QtCore.Qt.Key_Enter or event.key() == QtCore.Qt.Key_Return:
            self.update_display()

        if (QtCore.Qt.ControlModifier | QtCore.Qt.MetaModifier) & event.modifiers():
            super().keyPressEvent(event)
            return

        # The rest is to avoid editing suffix and prefix
        if len(event.text()) > 0:
            # Allow editing of the number or SI-prefix even if part of the prefix/suffix is selected.
            if self.lineEdit().selectedText():
                sel_start = self.lineEdit().selectionStart()
                sel_end = sel_start + len(self.lineEdit().selectedText())
                min_start = len(self.__prefix)
                max_end = len(self.__prefix) + len(self.cleanText())
                if sel_start < min_start:
                    sel_start = min_start
                if sel_end > max_end:
                    sel_end = max_end
                self.lineEdit().setSelection(sel_start, sel_end - sel_start)
            else:
                cursor_pos = self.lineEdit().cursorPosition()
                begin = len(self.__prefix)
                end = len(self.text()) - len(self.__suffix)
                if cursor_pos < begin:
                    self.lineEdit().setCursorPosition(begin)
                    return
                elif cursor_pos > end:
                    self.lineEdit().setCursorPosition(end)
                    return

        if event.key() == QtCore.Qt.Key_Left:
            if self.lineEdit().cursorPosition() == len(self.__prefix):
                return
        if event.key() == QtCore.Qt.Key_Right:
            if self.lineEdit().cursorPosition() == len(self.text()) - len(self.__suffix):
                return
        if event.key() == QtCore.Qt.Key_Home:
            self.lineEdit().setCursorPosition(len(self.__prefix))
            return
        if event.key() == QtCore.Qt.Key_End:
            self.lineEdit().setCursorPosition(len(self.text()) - len(self.__suffix))
            return

        super().keyPressEvent(event)

    def focusInEvent(self, event):
        super().focusInEvent(event)
        self.selectAll()
        return

    def focusOutEvent(self, event):
        self.update_display()
        super().focusOutEvent(event)
        return

    def wheelEvent(self, event):
        """
        Overwrites the wheel event. If the class variable disable_wheel = True, stepping with the mouse
        wheel is turned off and the wheel event is passed to the parent widget.

        Parameters
        ----------
        event : QWheelEvent
            A Qt QWheelEvent instance holding the wheel event information.
        """
        if self.disable_wheel:
            event.ignore()
        else:
            super().wheelEvent(event)

    def validate(self, text, position):
        """
        Access method to the validator. See IntegerValidator class for more information.

        Parameters
        ----------
        text : str
            String to be validated.
        position : int
            Current text cursor position.

        Returns
        -------
        (QValidator.State, str, int)
            - The returned validator state.
            - The input string.
            - The cursor position.
        """
        begin = len(self.__prefix)
        end = len(text) - len(self.__suffix)
        if position < begin:
            position = begin
        elif position > end:
            position = end

        if self.__prefix and text.startswith(self.__prefix):
            text = text[len(self.__prefix):]
        if self.__suffix and text.endswith(self.__suffix):
            text = text[:-len(self.__suffix)]

        state, string, position = self.validator.validate(text, position)

        text = self.__prefix + string + self.__suffix

        end = len(text) - len(self.__suffix)
        if position > end:
            position = end

        return state, text, position

    def fixup(self, text):
        """
        Takes an invalid string and tries to fix it in order to pass validation.
        The returned string is not guaranteed to pass validation.

        Parameters
        ----------
        text : str
            A string that has not passed validation and needs to be fixed.

        Returns
        -------
        str
            The resulting string from the fix attempt.
        """
        return self.validator.fixup(text)

    def valueFromText(self, text):
        """
        This method is responsible for converting a string displayed in the SpinBox into an integer value.
        The input string is already stripped of prefix and suffix. Only the SI-prefix may be present.

        Parameters
        ----------
        text : str
            The display string to be converted into a numeric value. This string must conform to the validator.

        Returns
        -------
        int
            The numeric value converted from the input string.
        """
        group_dict = self.validator.get_group_dict(text)
        if not group_dict:
            return False

        if not group_dict['mantissa']:
            return False

        si_prefix = group_dict['si']
        if si_prefix is None:
            si_prefix = ''
        si_scale = self._unit_prefix_dict[si_prefix.replace('u', 'µ')]

        unscaled_value = int(group_dict['mantissa'])
        if group_dict['exponent'] is not None:
            scale_factor = 10 ** int(group_dict['exponent'].replace('e', '').replace('E', ''))
            unscaled_value = unscaled_value * scale_factor

        value = unscaled_value * si_scale
        return value

    def textFromValue(self, value):
        """
        This method is responsible for mapping the underlying value to a string to display in the SpinBox.
        Suffix and Prefix are not handled here, only the SI-prefix.

        Parameters
        ----------
        value : int
            The numeric value to be formatted into a string.

        Returns
        -------
        str
            The formatted string representing the input value.
        """
        # Convert the integer value to a string
        sign = '-' if value < 0 else ''
        value_str = str(abs(value))

        # find out the index of the least significant non-zero digit
        for digit_index in range(len(value_str)):
            if value_str[digit_index:].count('0') == len(value_str) - digit_index:
                break

        # get the engineering notation exponent (multiple of 3)
        missing_zeros = (len(value_str) - digit_index) % 3
        exponent = len(value_str) - digit_index - missing_zeros

        # the scaled integer string that is still missing the order of magnitude (si-prefix or e)
        integer_str = value_str[:digit_index + missing_zeros]

        space = ' ' if self.__suffix else ''
        # Add si-prefix or, if the exponent is too big, add e-notation
        if 2 < exponent <= 24:
            si_prefix = ' ' + 'kMGTPEZY'[exponent // 3 - 1]
        elif exponent > 24:
            si_prefix = 'e{0:d}'.format(exponent) + space
        else:
            si_prefix = space

        # Assemble the string and return it
        return sign + integer_str + si_prefix

    def stepEnabled(self):
        """
        Enables stepping (mouse wheel, arrow up/down, clicking, PgUp/Down) by default.
        """
        return self.StepUpEnabled | self.StepDownEnabled

    def stepBy(self, steps):
        """
        This method increments the value of the SpinBox when the user triggers a step
        (by pressing PgUp/PgDown/Up/Down, MouseWheel movement, or clicking on the arrows).
        It handles cases where the new value to be set is out of bounds. The absolute value
        of a single step increment is also managed here to avoid accumulating rounding errors
        or discrepancies between self.value and the displayed text.

        Parameters
        ----------
        steps : int
            Number of steps to increment (NOT the absolute step size).
        """

        steps = int(steps)
        value = self.__value  # working copy of current value
        sign = -1 if steps < 0 else 1  # determine sign of step
        if self.dynamic_stepping:
            for i in range(abs(steps)):
                if value == 0:
                    step = max(1, self.__minimalStep)
                else:
                    integer_str = str(abs(value))
                    if len(integer_str) > 1:
                        step = 10 ** (len(integer_str) - 2)
                        # Handle the transition to lower order of magnitude
                        if integer_str.startswith('10') and (sign * value) < 0:
                            step = step // 10
                    else:
                        step = 1

                    step = max(step, self.__minimalStep)

                value += sign * step
        else:
            value = value + max(self.__minimalStep * steps, self.__singleStep * steps)

        self.setValue(value)
        return

    def selectAll(self):
        begin = len(self.__prefix)
        text = self.cleanText()
        if text.endswith(' '):
            selection_length = len(text) + 1
        elif len(text) > 0 and text[-1] in self._unit_prefix_dict:
            selection_length = len(text) - 1
        else:
            selection_length = len(text)
        self.lineEdit().setSelection(begin, selection_length)



# -*- coding: utf-8 -*-
"""
This file contains Qt models for Python data structures.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['DictTableModel', 'ListTableModel']

from PySide2 import QtCore
from typing import Any, Optional, Union, Sequence
from qudi.util.mutex import RecursiveMutex


class DictTableModel(QtCore.QAbstractTableModel):
    """Qt model storing a table in dictionaries.
    """
    def __init__(self, headers: Union[str, Sequence[str]]):
        super().__init__()
        self._lock = RecursiveMutex()
        if isinstance(headers, str):
            self._headers = [headers]
        elif not all(isinstance(h, str) for h in headers):
            raise TypeError('DictTableModel header entries must be str type.')
        else:
            self._headers = list(headers)
        self._storage = dict()

    def rowCount(self, parent: Optional[QtCore.QModelIndex] = None) -> int:
        """Returns the number of stored items (rows).
        """
        with self._lock:
            return len(self._storage)

    def columnCount(self, parent: Optional[QtCore.QModelIndex] = None) -> int:
        """Returns the number of data fields (columns).
        """
        with self._lock:
            return len(self._headers)

    def flags(self, index: Optional[QtCore.QModelIndex] = None) -> QtCore.Qt.ItemFlags:
        """Determines what can be done with the given indexed cell.
        """
        return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable

    def data(self, index: QtCore.QModelIndex, role: QtCore.Qt.ItemDataRole) -> Any:
        """Get data from model for a given cell. Data can have a role that affects display.
        Re-Implement in subclass in order to support anything else than the 2 default columns.
        """
        with self._lock:
            if index.isValid() and role == QtCore.Qt.DisplayRole:
                key = self.get_key_by_index(index.row())
                if index.column() == 0:
                    return key
                elif index.column() == 1:
                    return self._storage[key]
            return None

    def headerData(self, section: int, orientation: QtCore.Qt.Orientation,
                   role: Optional[QtCore.Qt.ItemDataRole] = QtCore.Qt.DisplayRole) -> Any:
        """Data for the table view headers."""
        with self._lock:
            if role == QtCore.Qt.DisplayRole:
                if orientation == QtCore.Qt.Horizontal:
                    return self._headers[section]
            return None

    def get_key_by_index(self, n: int) -> Any:
        """Get a dict key by index number."""
        with self._lock:
            it = iter(self._storage)
            try:
                for _ in range(n+1):
                    key = next(it)
            except StopIteration:
                raise IndexError(f'Index {n:d} out of bounds for table model with '
                                 f'{len(self._storage):d} rows') from None
            return key

    def get_index_by_key(self, key: Any) -> int:
        """Get row index for dict key.

        Warning: Row index for a key changes when keys with lower index are removed.
        """
        with self._lock:
            for i, storage_key in enumerate(self._storage.keys()):
                if key == storage_key:
                    return i
            raise KeyError

    def __setitem__(self, key, value):
        with self._lock:
            if key in self._storage:
                row = self.get_index_by_key(key)
                self._storage[key] = value
                index = self.index(row, 1)
                self.dataChanged.emit(index, index)
            else:
                row = len(self._storage)
                self.beginInsertRows(QtCore.QModelIndex(), row, row)
                self._storage[key] = value
                self.endInsertRows()

    def __getitem__(self, item):
        with self._lock:
            return self._storage.__getitem__(item)

    def __repr__(self):
        with self._lock:
            return repr(self._storage)

    def __len__(self):
        with self._lock:
            return len(self._storage)

    def __delitem__(self, key):
        with self._lock:
            if key not in self._storage:
                raise KeyError
            self.pop(key)

    def __iter__(self):
        with self._lock:
            return self._storage.__iter__()

    def __contains__(self, item):
        with self._lock:
            return self._storage.__contains__(item)

    def clear(self):
        with self._lock:
            self.beginResetModel()
            self._storage.clear()
            self.endResetModel()

    def copy(self):
        model = DictTableModel(self._headers.copy())
        model._storage = self._storage.copy()
        return model

    def update(self, *args, **kwargs):
        with self._lock:
            update_dict = dict(*args, **kwargs)
            for key, value in update_dict.items():
                self.__setitem__(key, value)

    def pop(self, *args):
        """
        Remove a key from a dictionary and return its value.

        Parameters
        ----------
        *args : tuple
            The arguments should include:
            - key : The key to remove from the dictionary.
            - default : optional
                The value to return if the key is not found in the dictionary.

        Returns
        -------
        value
            The value associated with the removed key. If the key is not found and a default 
            value is provided, the default value is returned. Otherwise, a `KeyError` is raised.
        """
        with self._lock:
            if args[0] in self._storage:
                row = self.get_index_by_key(args[0])
                self.beginRemoveRows(QtCore.QModelIndex(), row, row)
                ret = self._storage.pop(args[0])
                self.endRemoveRows()
                return ret
            elif len(args) > 1:
                return args[1]

    def get(self, *args):
        """
        Get the value associated with a key from a dictionary.

        Parameters
        ----------
        *args : tuple
            The arguments should include:
            - key : The key to look up in the dictionary.
            - default : optional
                The value to return if the key is not found in the dictionary.

        Returns
        -------
        value
            The value associated with the key in the dictionary. If the key is not found 
            and a default value is provided, the default value is returned.
        """
        with self._lock:
            return self._storage.get(*args)

    def values(self):
        with self._lock:
            return self._storage.values()

    def keys(self):
        with self._lock:
            return self._storage.keys()

    def items(self):
        with self._lock:
            return self._storage.items()


class ListTableModel(QtCore.QAbstractTableModel):
    """Qt model storing a table in lists.
    """

    def __init__(self, headers: Union[str, Sequence[str]]):
        super().__init__()
        self._lock = RecursiveMutex()
        if isinstance(headers, str):
            self._headers = headers
        elif not all(isinstance(h, str) for h in headers):
            raise TypeError('DictTableModel header entries must be str type.')
        else:
            self._headers = list(headers)
        self._storage = list()

    def rowCount(self, parent: Optional[QtCore.QModelIndex] = None):
        """Gives the number of stored items (rows).
        """
        with self._lock:
            return len(self._storage)

    def columnCount(self, parent: Optional[QtCore.QModelIndex] = None):
        """Gives the number of data fields (columns).
        """
        return len(self._headers)

    def flags(self, index: QtCore.QModelIndex) -> QtCore.Qt.ItemFlags:
        """Determines what can be done with entry cells in the table view.
        """
        return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable

    def data(self, index: QtCore.QModelIndex, role: QtCore.Qt.ItemDataRole) -> Any:
        """Get data from model for a given cell. Data can have a role that affects display.
        """
        with self._lock:
            if index.isValid() and role == QtCore.Qt.DisplayRole:
                return self._storage[index.row()]
            return None

    def headerData(self, section: int, orientation: QtCore.Qt.Orientation,
                   role: Optional[QtCore.Qt.ItemDataRole] = QtCore.Qt.DisplayRole) -> Any:
        """Data for the table view headers.
        """
        if role == QtCore.Qt.DisplayRole:
            if orientation == QtCore.Qt.Horizontal:
                return self._headers[section]
        return None

    def __setitem__(self, key, value):
        with self._lock:
            if key < 0:
                key = key + len(self._storage)
            self._storage[key] = value
            self.dataChanged.emit(self.index(key, 0), self.index(key, len(self._headers) - 1))

    def __getitem__(self, key):
        with self._lock:
            return self._storage[key]

    def __repr__(self):
        with self._lock:
            return repr(self._storage)

    def __len__(self):
        with self._lock:
            return len(self._storage)

    def __delitem__(self, key):
        with self._lock:
            if key < 0:
                key = key + len(self._storage)
            self.pop(key)

    def __iter__(self):
        with self._lock:
            return iter(self._storage)

    def __contains__(self, item):
        with self._lock:
            return item in self._storage

    def insert(self, n, data):
        """
        Insert a row into a table before the nth element.

        Parameters
        ----------
        n : int
            Index before which the row should be inserted.
        data : object
            Row to insert into the table.

        """
        with self._lock:
            if 0 <= n <= len(self._storage):
                self.beginInsertRows(QtCore.QModelIndex(), n, n)
                self._storage.insert(n, data)
                self.endInsertRows()
            else:
                raise IndexError

    def append(self, data):
        """
        Append a row to a table.

        Parameters
        ----------
        data : object
            Row to append to the table.

        """
        with self._lock:
            n = len(self._storage)
            self.beginInsertRows(QtCore.QModelIndex(), n, n)
            self._storage.append(data)
            self.endInsertRows()

    def pop(self, n):
        """
        Remove the nth row from a table.

        Parameters
        ----------
        n : int
            Index of the row to remove.

        Returns
        -------
        object
            Removed row.

        """
        with self._lock:
            if 0 <= n < len(self._storage):
                self.beginRemoveRows(QtCore.QModelIndex(), n, n)
                ret = self._storage.pop(n)
                self.endRemoveRows()
                return ret
            else:
                raise IndexError

    def extend(self, seq):
        with self._lock:
            seq = list(seq)
            n = len(self._storage)
            m = n + len(seq)
            self.beginInsertRows(QtCore.QModelIndex(), n, m)
            self._storage.extend(seq)
            self.endInsertRows()

    def remove(self, value):
        with self._lock:
            row = self._storage.index(value)
            self.pop(row)

    def count(self, value):
        with self._lock:
            return self._storage.count(value)

    def copy(self):
        model = ListTableModel(self._headers)
        model._storage = self._storage.copy()
        return model



# -*- coding: utf-8 -*-
"""

"""

__all__ = ['ModuleFinder', 'QudiModules']

import importlib
import inspect
import logging
from typing import List, Type, Dict, Iterable

from qudi.core import Connector, ConfigOption, Base, LogicBase, GuiBase
from qudi.util.helpers import iter_modules_recursive


log = logging.getLogger(__package__)


class ModuleFinder:
    """
    """
    @staticmethod
    def is_qudi_module(obj: type) -> bool:
        if obj in [Base, LogicBase, GuiBase]:
            return False
        return inspect.isclass(obj) and issubclass(obj, Base) and not inspect.isabstract(obj)

    @staticmethod
    def get_module_names_from_ns(namespace: object) -> List[str]:
        module_names = [mod_finder.name for mod_finder in
                        iter_modules_recursive(namespace.__path__, f'{namespace.__name__}.')]
        # Remove duplicates
        return list(dict.fromkeys(module_names))

    @classmethod
    def get_qudi_classes_in_module(cls, module: object) -> Dict[str, Type[Base]]:
        members = inspect.getmembers(module, cls.is_qudi_module)
        return {f'{module.__name__}.{name}': obj for name, obj in members if
                obj.__module__ == module.__name__}

    @classmethod
    def get_qudi_modules_from_ns(cls, namespace: object) -> Dict[str, Type[Base]]:
        qudi_modules = dict()
        for module_name in cls.get_module_names_from_ns(namespace):
            try:
                module = importlib.import_module(module_name)
            except:
                log.warning(f'Error during import of module "{module_name}"')
                continue
            qudi_modules.update(cls.get_qudi_classes_in_module(module))
        return qudi_modules

    @classmethod
    def get_qudi_modules(cls) -> Dict[str, Dict[str, Type[Base]]]:
        try:
            import qudi.gui as _gui_ns
        except ImportError:
            _gui_ns = None
        try:
            import qudi.logic as _logic_ns
        except ImportError:
            _logic_ns = None
        try:
            import qudi.hardware as _hardware_ns
        except ImportError:
            _hardware_ns = None
        modules = dict()
        if _gui_ns is not None:
            modules.update(cls.get_qudi_modules_from_ns(_gui_ns))
        if _logic_ns is not None:
            modules.update(cls.get_qudi_modules_from_ns(_logic_ns))
        if _hardware_ns is not None:
            modules.update(cls.get_qudi_modules_from_ns(_hardware_ns))
        return modules


class QudiModules:
    """
    """

    def __init__(self):
        # import all qudi module classes if possible (log all errors upon import)
        self._qudi_modules = {mod[5:] if mod.startswith('qudi.') else mod: cls for mod, cls in
                              ModuleFinder.get_qudi_modules().items()}
        # Collect all connectors for all modules
        self._module_connectors = {
            mod: list(cls._meta['connectors'].values()) for mod, cls in self._qudi_modules.items()
        }
        # Get for each connector in each module compatible modules to connect to
        self._module_connectors_compatible_modules = {
            mod: self._modules_for_connectors(conn) for mod, conn in self._module_connectors.items()
        }
        # Get all ConfigOptions for all modules
        self._module_config_options = {
            mod: list(cls._meta['config_options'].values()) for mod, cls in
            self._qudi_modules.items()
        }

    def _modules_for_connectors(self, connectors: Iterable[Connector]) -> Dict[str, List[str]]:
        return {conn.name: self._modules_for_connector(conn) for conn in connectors}

    def _modules_for_connector(self, connector: Connector) -> List[str]:
        interface = connector.interface
        bases = {mod: {c.__name__ for c in cls.mro()} for mod, cls in self._qudi_modules.items()}
        return list(mod for mod, base_names in bases.items() if interface in base_names)

    @property
    def available_modules(self) -> List[str]:
        return list(self._qudi_modules)

    def module_connectors(self, module: str) -> List[Connector]:
        return self._module_connectors[module].copy()

    def module_connector_targets(self, module: str) -> Dict[str, List[str]]:
        return {k: v.copy() for k, v in self._module_connectors_compatible_modules[module].items()}

    def module_config_options(self, module: str) -> List[ConfigOption]:
        return self._module_config_options[module].copy()
